from __future__ import annotations

import logging
import typing

import angr
import claripy
import cle

from ... import exceptions, platforms, utils
from .. import emulator
from .default import configure_default_plugins, configure_default_strategy
from .factory import PatchedObjectFactory
from .machdefs import AngrMachineDef
from .simos import HookableSimOS

log = logging.getLogger(__name__)


class HookHandler(angr.SimProcedure):
    """SimProcedure for implementing "finish" hooks.

    This requires a callback as an extra kwarg.
    """

    def run(self, *args, callback, parent):
        emu = ConcreteAngrEmulator(self.state, parent)
        callback(emu)
        return None


class AngrEmulator(
    emulator.Emulator,
    emulator.InstructionHookable,
    emulator.FunctionHookable,
    emulator.SyscallHookable,
    emulator.MemoryReadHookable,
    emulator.MemoryWriteHookable,
    emulator.ConstrainedEmulator,
):
    """
    Angr symbolic execution emulator

    This is primarily designed to support symbolic execution,
    although subclasses can configure angr however they like.

    One challenge with symbolic execution is that
    it doesn't work on a single machine state,
    but rather multiple machine states representing
    parallel execution paths.

    As such, this interface doesn't yet fully support
    all features of the base Emulator class;
    it's not clear what reading or writing machine state
    means when there's more than one state.
    """

    PAGE_SIZE = 4096

    name = "angr-emulator"
    description = "an emulator using angr as its backend"
    version = "0.0"

    def __init__(self, platform: platforms.Platform, preinit=None, init=None):
        # Initializaed bit; tells us if angr state is initialized
        self._initialized: bool = False

        # Dirty bit; tells us if we can modify self.state
        self._dirty: bool = False

        # Linear mode bit; tells us if we're running in forced linear execution
        self._linear: bool = False

        # Plugin preset; tells us which plugin preset to use.
        self._plugin_preset = "default"

        # The platform definition;
        # Holds global info about the platform
        self.platdef: platforms.PlatformDef = platforms.PlatformDef.for_platform(
            platform
        )

        # The machine definition;
        # helps translate between angr and smallworld
        self.machdef: AngrMachineDef = AngrMachineDef.for_platform(platform)

        self.platform = platform
        self.preinit = preinit
        self.init = init

        # Cached state modifications,
        # to be applied during initialization
        self._code: typing.List[typing.Tuple[int, bytes]] = list()
        self._register_contents: typing.Dict[
            str, typing.Union[int, claripy.ast.bv.BV]
        ] = dict()
        self._register_labels: typing.Dict[str, str] = dict()
        self._memory_maps: typing.List[typing.Tuple[int, int]] = list()
        self._memory_contents: typing.List[
            typing.Tuple[int, typing.Union[bytes, claripy.ast.bv.BV]]
        ] = list()
        self._memory_labels: typing.List[typing.Tuple[int, int, str]] = list()
        self._instr_hooks: typing.List[
            typing.Tuple[int, typing.Callable[[emulator.Emulator], None]]
        ] = list()
        self._gb_instr_hook: typing.Optional[
            typing.Callable[[emulator.Emulator], None]
        ] = None
        self._func_hooks: typing.List[
            typing.Tuple[int, typing.Callable[[emulator.Emulator], None]]
        ] = list()
        self._syscall_hooks: typing.Dict[
            int, typing.Callable[[emulator.Emulator], None]
        ] = dict()
        self._gb_syscall_hook: typing.Optional[
            typing.Callable[[emulator.Emulator, int], None]
        ] = None
        self._read_hooks: typing.List[
            typing.Tuple[
                int,
                int,
                typing.Callable[
                    [emulator.Emulator, int, int, claripy.ast.bv.BV],
                    typing.Optional[claripy.ast.bv.BV],
                ],
            ]
        ] = list()
        self._gb_read_hook: typing.Optional[
            typing.Callable[
                [emulator.Emulator, int, int, claripy.ast.bv.BV],
                typing.Optional[claripy.ast.bv.BV],
            ]
        ] = None
        self._write_hooks: typing.List[
            typing.Tuple[
                int,
                int,
                typing.Callable[[emulator.Emulator, int, int, claripy.ast.bv.BV], None],
            ]
        ] = list()
        self._gb_write_hook: typing.Optional[
            typing.Callable[[emulator.Emulator, int, int, claripy.ast.bv.BV], None]
        ] = None
        self._code_bounds: typing.List[typing.Tuple[int, int]] = list()
        self._exit_points: typing.Set[int] = set()
        self._constraints: typing.List[claripy.ast.bool.Bool] = list()
        self._extensions: typing.Dict[str, typing.Any] = dict()

    def initialize(self):
        """Initialize the emulator

        To take advantage of CLE, we need to know about all code
        before we initialize the angr state objects.
        However, applying register and memory changes requires access
        to the angr state object.

        SmallWorld doesn't support ordering how state is applied,
        so I need to collect it, then apply it once emulation starts.

        This function is invoked automatically when you cycle the emulator,
        but you're free to invoke it early if you want.
        """

        if self._initialized:
            raise exceptions.ConfigurationError("Cannot reinitialize running emulator")

        # Build a single, coherent byte stream out of our code
        self._code.sort(key=lambda x: x[0])

        code = utils.SparseIO()
        segments = list()

        if len(self._code) > 0:
            code_ranges = utils.RangeCollection()
            base_address = self._code[0][0]
            for addr, data in self._code:
                size = len(data)
                code.seek(addr - base_address)
                code.write(data)
                code_ranges.add_range((addr, addr + size))
            segments = list(
                map(lambda x: (x[0] - base_address, x[0], x[1] - x[0]), code_ranges)
            )

        # Create an angr project using a blank byte stream,
        # and registered as self-modifying so we can load more code later.
        backend = cle.Blob(
            None, code, is_main_bin=True, arch=self.machdef.angr_arch, segments=segments
        )
        loader = cle.Loader(backend)
        self.proj: angr.Project = angr.Project(
            loader,
            engine=self.machdef.angr_engine,
            selfmodifying_code=True,
            simos=HookableSimOS,
        )
        # Override the default factory
        self.proj.factory = PatchedObjectFactory(
            self.proj, type(self.proj.factory.default_engine)
        )

        # Configure default plugin preset.
        # Do this before creating the state, so the state inherits the correct preset.
        # Also do this before preinit, so that analyses can inherit from the default.
        configure_default_plugins(self)

        # If preinit is specified, run it.
        if self.preinit is not None:
            self.preinit(self)

        # Create a completely blank entry state
        self.state = self.proj.factory.blank_state(
            plugin_preset=self._plugin_preset,
            add_options={
                # angr.options.BYPASS_UNSUPPORTED_SYSCALL,
                angr.options.SYMBOL_FILL_UNCONSTRAINED_REGISTERS,
                angr.options.SYMBOL_FILL_UNCONSTRAINED_MEMORY,
            },
            remove_options={
                angr.options.SIMPLIFY_CONSTRAINTS,
                angr.options.SIMPLIFY_EXIT_GUARD,
                angr.options.SIMPLIFY_EXIT_STATE,
                angr.options.SIMPLIFY_EXIT_TARGET,
                angr.options.SIMPLIFY_EXPRS,
                angr.options.SIMPLIFY_MEMORY_READS,
                angr.options.SIMPLIFY_MEMORY_WRITES,
                angr.options.SIMPLIFY_REGISTER_READS,
                angr.options.SIMPLIFY_REGISTER_WRITES,
            },
        )

        # Create a simulation manager for our entry state
        self.mgr = self.proj.factory.simulation_manager(self.state, save_unsat=True)

        # Configure default simulation strategy.
        configure_default_strategy(self)

        # If we have an init runner, run it.
        if self.init:
            self.init(self)

        # Mark this emulator as initialized
        self._initialized = True

        # Read back any changes that happened before initialization.
        for name, content in self._register_contents.items():
            self.write_register_content(name, content)

        for name, label in self._register_labels.items():
            self.write_register_label(name, label)

        for addr, size in self._memory_maps:
            self.map_memory(addr, size)

        for addr, content in self._memory_contents:
            self.write_memory_content(addr, content)

        self._write_memory_label_bulk(self._memory_labels)

        for addr, func in self._instr_hooks:
            self.hook_instruction(addr, func)

        if self._gb_instr_hook is not None:
            self.hook_instructions(self._gb_instr_hook)

        for addr, func in self._func_hooks:
            self.hook_function(addr, func)

        for num, func in self._syscall_hooks.items():
            self.hook_syscall(num, func)

        if self._gb_syscall_hook is not None:
            self.hook_syscalls(self._gb_syscall_hook)

        for start, end, func in self._read_hooks:
            self.hook_memory_read_symbolic(start, end, func)

        if self._gb_read_hook is not None:
            self.hook_memory_reads_symbolic(self._gb_read_hook)

        for start, end, func in self._write_hooks:
            self.hook_memory_write_symbolic(start, end, func)

        if self._gb_write_hook is not None:
            self.hook_memory_writes_symbolic(self._gb_write_hook)

        for start, end in self._code_bounds:
            self.add_bound(start, end)

        for addr in self._exit_points:
            self.add_exit_point(addr)

        for expr in self._constraints:
            self.add_constraint(expr)

        self.state.scratch.extensions |= self._extensions

    def read_register_symbolic(self, name: str) -> claripy.ast.bv.BV:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Cannot read registers before initialization"
            )

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Reading registers not supported once execution begins."
            )

        if name == "pc":
            name = self.platdef.pc_register

        (off, size) = self.machdef.angr_reg(name)
        return self.state.registers.load(off, size)

    def read_register_content(self, name: str) -> int:
        out = self.read_register_symbolic(name)
        if out.symbolic:
            raise exceptions.SymbolicValueError(f"Register {name} is symbolic")
        else:
            return out.concrete_value

    def read_register_type(self, name: str) -> typing.Optional[typing.Any]:
        return None

    def read_register_label(self, name: str) -> typing.Optional[str]:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Cannot read registers before initialization"
            )

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Reading register not supported once execution begins."
            )

        if name == "pc":
            name = self.platdef.pc_register

        try:
            # This considers all BVSs to be labeled values;
            # if it has a name, we're giving it to you.
            (off, size) = self.machdef.angr_reg(name)
            out = self.state.registers.load(off, size)
            if out.symbolic:
                if out.op == "BVS":
                    # This is a "pure" label; we can return it.
                    return out.args[0]
                else:
                    # This is a mixed expression; we can't return it
                    log.warn(f"Register {name} contains a symbolic expression")
                    raise exceptions.SymbolicValueError(
                        f"Register {name} contains a symbolic expression"
                    )
            else:
                # No propagated label
                return None
        except ValueError:
            # TODO: Handle invalid registers more gracefully
            return None

    def write_register_content(
        self, name: str, content: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        if not self._initialized and content is not None:
            if name == "pc":
                name = self.platdef.pc_register
            # Test that the angr register exists
            _, _ = self.machdef.angr_reg(name)
            self._register_contents[name] = content
            return
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Writing registers not supported once execution begins."
            )
        # This will replace any value - symbolic or otherwise - currently in the emulator.
        # Since labels are treated as symbolic values, it must be called before
        # write_register_label().

        if name == "pc":
            name = self.platdef.pc_register
        (off, size) = self.machdef.angr_reg(name)

        if content is None:
            v = claripy.BVS("UNINITIALIZED", size * 8)
        elif isinstance(content, int):
            v = claripy.BVV(content, size * 8)
        elif isinstance(content, claripy.ast.bv.BV):
            v = content
        else:
            raise TypeError(
                f"Content for register {name} was {type(content)}; expected (None | int | claripy.ast.bv.BV)"
            )

        if v.size() != size * 8:
            raise exceptions.ConfigurationError(
                f"Register {name} has size {size}, but content has size {v.size() // 8}"
            )

        self.state.registers.store(off, v)

    def write_register_type(
        self, name: str, type: typing.Optional[typing.Any] = None
    ) -> None:
        pass

    def write_register_label(
        self, name: str, label: typing.Optional[str] = None
    ) -> None:
        if label is None:
            return
        elif not self._initialized:
            if name == "pc":
                name = self.platdef.pc_register
            # Test that the angr register exists
            _, _ = self.machdef.angr_reg(name)
            self._register_labels[name] = label
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Writing registers not supported once execution begins."
            )
        else:
            if name == "pc":
                name = self.platdef.pc_register
            (off, size) = self.machdef.angr_reg(name)

            # This will bind whatever value is currently in the register
            # to a symbol named after the label
            # The same label will ALWAYS map to the same symbol!
            # This can introduce unintended restrictions on exploration,
            # or even a contradictory state.
            s = claripy.BVS(label, size * 8, explicit_name=True)
            v = self.state.registers.load(off, size)

            if not self.state.solver.satisfiable(extra_constraints=[v == s]):
                # Creating this binding will definitely cause a contradiction.
                # Have you already used this label somewhere else?
                raise exceptions.ConfigurationError(
                    f"Contradiction binding register {name} to label {label}"
                )

            # Passing the last check doesn't guarantee you're safe.
            # There may be over-constraints.  Please be careful.
            self.state.registers.store(off, s)
            self.state.solver.add(v == s)

    def read_memory_content(self, address: int, size: int) -> bytes:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Cannot read memory before initialization"
            )

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Reading memory not supported once execution begins."
            )
        v = self.state.memory.load(address, size)
        if v.symbolic:
            log.warn(f"Memory at {hex(address)} ({size} bytes) is symbolic")
            raise exceptions.SymbolicValueError(f"Memory at {hex(address)} is symbolic")

        # Annoyingly, there isn't an easy way to convert BVV to bytes.
        return bytes([v.get_byte(i).concrete_value for i in range(0, size)])

    def read_memory_symbolic(self, address: int, size: int) -> claripy.ast.bv.BV:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Cannot read memory before initialization"
            )

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Reading memory not supported once execution begins."
            )
        return self.state.memory.load(address, size)

    def read_memory_type(self, address: int, size: int) -> typing.Optional[typing.Any]:
        return None

    def read_memory_label(self, address: int, size: int) -> typing.Optional[str]:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Cannot read memory before initialization"
            )

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Writing memory not supported once execution begins."
            )
        v = self.state.memory.load(address, size)
        if v.symbolic():
            if v.op == "Extract":
                # You got a piece of a possibly-labeled expression
                # Try parsing the inner expression to see if it's a single symbol.
                v = v.args[2]

            if v.op == "BVS":
                # You got a single symbol; I'll treat it as a label
                return v.args[0]
            else:
                # You got a symbolic expression; I can't decode it further
                log.warn(f"Memory at {hex(address)} ({size} bytes) is symbolic: {v}")
                raise exceptions.SymbolicValueError(
                    f"Memory at {hex(address)} is symbolic"
                )
        else:
            # Definitely no labels here
            return None

    def map_memory(self, address: int, size: int) -> None:
        if not self._initialized:
            self._memory_maps.append((address, size))
        else:
            if self._dirty and not self._linear:
                raise NotImplementedError(
                    "Mapping memory not supported once execution begins."
                )
            # Unlike Unicorn, angr doesn't care about pages.
            region = (address, address + size)
            self.state.scratch.memory_map.add_range(region)

    def get_memory_map(self) -> typing.List[typing.Tuple[int, int]]:
        if not self._initialized:
            return list(map(lambda x: (x[0], x[0] + x[1]), self._memory_maps))
        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Mapping memory not supported once execution begins."
            )
        return list(self.state.scratch.memory_map.ranges)

    def write_memory_content(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ) -> None:
        if not self._initialized:
            self._memory_contents.append((address, content))
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Writing memory not supported once execution begins."
            )
        else:
            log.info(f"Storing {len(content)} bytes at {hex(address)}")
            if isinstance(content, bytes):
                content = claripy.BVV(content)
            self.state.memory.store(address, content, inspect=False)

    def write_memory_type(
        self, address: int, size: int, type: typing.Optional[typing.Any] = None
    ) -> None:
        pass

    def write_memory_label(
        self, address: int, size: int, label: typing.Optional[str] = None
    ) -> None:
        if label is None:
            return
        elif not self._initialized:
            self._memory_labels.append((address, size, label))
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Writing memory not supported once execution begins."
            )
        else:
            # This will bind whatever value is currently at this address
            # to a symbol named after the label
            #
            # The same label will ALWAYS map to the same symbol!
            # This can introduce unintended restrictions on exploration,
            # or even a contradictory state.
            #
            # This will trigger angr's default value computation,
            # which may cause spurious results in some analyses.
            s = claripy.BVS(label, size * 8, explicit_name=True)
            v = self.state.memory.load(address, size)
            if not self.state.solver.satisfiable(extra_constraints=[v == s]):
                # Creating this binding will definitely cause a contradiction.
                # Have you already used this label somewhere else?
                raise exceptions.ConfigurationError(
                    f"Contradiction binding memory at {hex(address)} to label {label}"
                )

            # Passing the last check doesn't mean you're safe.
            # There may be over-constraints.  Please be careful.
            self.state.memory.store(address, s, inspect=False)
            self.state.solver.add(v == s)

    def _write_memory_label_bulk(
        self, labels: typing.List[typing.Tuple[int, int, typing.Optional[str]]]
    ) -> None:
        # Variant of write_memory_label that applies multiple labels in one go.
        # This addresses a performance bottleneck during initialization;
        # it's not really intended for public use,
        # as it lacks some of the safety checks.
        cs = list()
        for addr, size, label in labels:
            if label is None:
                continue
            s = claripy.BVS(label, size * 8, explicit_name=True)
            v = self.state.memory.load(addr, size)
            self.state.memory.store(addr, s)
            cs.append(s == v)
        c = claripy.And(*cs)
        self.state.solver.add(c)

    def write_code(self, address: int, content: bytes):
        if self._initialized:
            self.write_memory_content(address, content)
        else:
            self._code.append((address, content))

    def hook_instruction(
        self, address: int, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        if not self._initialized:
            self._instr_hooks.append((address, function))

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Instruction hooking not supported once execution begins"
            )

        elif address in self.state.scratch.insn_bps:
            raise exceptions.ConfigurationError(
                "Instruction at address {hex(address)} is already hooked"
            )

        else:

            def hook_handler(state):
                emu = ConcreteAngrEmulator(state, self)
                function(emu)

                # An update to angr means some operations on `state`
                # will clobber this variable, which causes bad problems.
                # Unclobber it, just in case
                state.inspect.action_attrs_set = True

            bp = self.state.inspect.b(
                "instruction",
                when=angr.BP_BEFORE,
                action=hook_handler,
                instruction=address,
            )
            self.state.scratch.insn_bps[address] = bp

    def unhook_instruction(self, address: int) -> None:
        if not self._initialized:
            self._instr_hooks = list(
                filter(lambda x: x[0] != address, self._instr_hooks)
            )

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Instruction hooking not supported once execution begins"
            )

        elif address not in self.state.scratch.insn_bps:
            raise exceptions.ConfigurationError(
                "Instruction at address {hex(address)} is not hooked"
            )

        else:
            bp = self.state.scratch.insn_bps[address]
            del self.state.scratch.insn_bps[address]

            self.state.inspect.remove_breakpoint("instruction", bp)

    def hook_instructions(
        self, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        if not self._initialized:
            self._gb_instr_hook = function

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Global instruction hooking not supported once execution begins"
            )

        elif self.state.scratch.global_insn_bp is not None:
            raise exceptions.ConfigurationError(
                "Global instruction hook already registered"
            )

        else:

            def hook_handler(state):
                emu = ConcreteAngrEmulator(state, self)
                function(emu)

                # An update to angr means some operations on `state`
                # will clobber this variable, which causes bad problems.
                # Unclobber it, just in case
                state.inspect.action_attrs_set = True

            self.state.scratch.global_insn_bp = self.state.inspect.b(
                "instruction", when=angr.BP_BEFORE, action=hook_handler
            )

    def unhook_instructions(self) -> None:
        if not self._initialized:
            self._gb_instruction_hook = None

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Global instruction unhooking not supported once execution begins"
            )

        elif self.state.scratch.global_insn_bp is None:
            raise exceptions.ConfigurationError("No global instruction hook present")

        else:
            bp = self.state.scratch.global_insn_bp
            self.state.scratch.global_insn_bp = None
            self.state.inspect.remove_breakpoint("instruction", bp)

    def hook_function(
        self, address: int, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        if not self._initialized:
            self._func_hooks.append((address, function))

        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot hook functions once emulation starts")

        else:
            hook = HookHandler(callback=function, parent=self)
            self.proj.hook(address, hook, 0)

            self.map_memory(address, 1)
            self.state.scratch.func_bps[address] = None

    def unhook_function(self, address: int) -> None:
        if not self._initialized:
            self._func_hooks = list(filter(lambda x: x[0] != address, self._func_hooks))

        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot unhook functions once emulation starts")

        else:
            self.proj.unhook(address)

    def hook_syscall(
        self, number: int, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        if not self._initialized:
            self._syscall_hooks[number] = function
        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot hook syscalls once emulation starts")
        elif number in self.state.scratch.syscall_funcs:
            raise exceptions.ConfigurationError(
                f"Already have a syscall hook for {number}"
            )
        else:

            def syscall_handler(state):
                function(ConcreteAngrEmulator(state, self))

            self.state.scratch.syscall_funcs[number] = syscall_handler

    def unhook_syscall(self, number: int) -> None:
        if not self._initialized:
            del self._syscall_hooks[number]

        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot unhook syscalls once emulation starts")

        elif number not in self.state.scratch.syscall_funcs:
            raise exceptions.ConfigurationError(f"No syscall hook for {number}")

        else:
            del self.state.scratch.syscall_funcs[number]

    def hook_syscalls(
        self, function: typing.Callable[[emulator.Emulator, int], None]
    ) -> None:
        if not self._initialized:
            self._gb_syscall_hook = function

        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot hook syscalls once emulation starts")

        elif self.state.scratch.global_syscall_func is not None:
            raise exceptions.ConfigurationError("Already have a global syscall hook")

        else:

            def syscall_handler(state, number: int):
                function(ConcreteAngrEmulator(state, self), number)

            self.state.scratch.global_syscall_func = syscall_handler

    def unhook_syscalls(self) -> None:
        if self._dirty and not self._linear:
            raise NotImplementedError("Cannot unhook syscalls once emulation starts")

        if self.state.scratch.global_syscall_func is None:
            raise exceptions.ConfigurationError("No global syscall hook registered")
        self.state.scratch.global_syscall_func = None

    def hook_memory_read_symbolic(
        self,
        start: int,
        end: int,
        function: typing.Callable[
            [emulator.Emulator, int, int, claripy.ast.bv.BV],
            typing.Optional[claripy.ast.bv.BV],
        ],
    ) -> None:
        if not self._initialized:
            self._read_hooks.append((start, end, function))

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory hooking not supported once execution begins"
            )

        elif (start, end) in self.state.scratch.mem_read_bps:
            raise exceptions.ConfigurationError(
                f"{hex(start)}-{hex(end)} already hooked for reads"
            )

        else:
            # This uses angr's conditional breakpoint facility
            def read_condition(state):
                # The breakpoint condition.
                # This needs to be a bit clever to detect reads at bound symbolic addresses.
                # The actual address won't get concretized until later
                read_start = state.inspect.mem_read_address
                if not isinstance(read_start, int):
                    if read_start.symbolic:
                        try:
                            values = state.solver.eval_atmost(read_start, 1)
                            # Truly unbound address.
                            # Assume it won't collapse to our hook address
                            if len(values) < 1:
                                return False
                            read_start = values[0]
                        except angr.errors.SimUnsatError:
                            return False
                        except angr.errors.SimValueError:
                            return False
                    else:
                        read_start = read_start.concrete_value
                    state.inspect.mem_read_address = read_start
                read_size = state.inspect.mem_read_length

                if read_size is None:
                    return False
                read_end = read_start + read_size

                rng = range(start, end)
                access_rng = range(read_start, read_end)
                return (
                    read_start in rng
                    or read_end - 1 in rng
                    or start in access_rng
                    or end - 1 in access_rng
                )

            def read_callback(state):
                # The breakpoint action.
                addr = state.inspect.mem_read_address
                size = state.inspect.mem_read_length
                expr = state.inspect.mem_read_expr
                res = function(ConcreteAngrEmulator(state, self), addr, size, expr)

                if res is None:
                    res = expr
                state.inspect.mem_read_expr = res

                # An update to angr means some operations on `state`
                # will clobber this variable, which causes bad problems.
                # Unclobber it, just in case
                state.inspect.action_attrs_set = True

            bp = self.state.inspect.b(
                "mem_read",
                when=angr.BP_AFTER,
                condition=read_condition,
                action=read_callback,
            )
            self.state.scratch.mem_read_bps[(start, end)] = bp

    def hook_memory_read(
        self,
        start: int,
        end: int,
        function: typing.Callable[
            [emulator.Emulator, int, int, bytes], typing.Optional[bytes]
        ],
    ):
        # Most of the machinery is handled in hook_memory_read_symbolic
        # We just need to concretize the inputs and claripy the outputs
        def sym_callback(
            emu: emulator.Emulator, addr: int, size: int, expr: claripy.ast.bv.BV
        ) -> typing.Optional[claripy.ast.bv.BV]:
            if not isinstance(emu, ConcreteAngrEmulator):
                raise TypeError("Where'd you get your emu?")
            if expr.symbolic:
                # Have the solver handle binding resolution for us.
                try:
                    values = emu.state.solver.eval_atmost(expr, 1)
                    if len(values) < 1:
                        raise exceptions.AnalysisError(f"No possible values fpr {expr}")
                    value = values[0].to_bytes(size, byteorder=self.byteorder)
                    log.info("Collapsed symbolic {expr} to {values[0]:x} for MMIO")
                except angr.errors.SimUnsatError:
                    raise exceptions.AnalysisError(f"No possible values for {expr}")
                except angr.errors.SimValueError:
                    value = b"\x00" * size
            else:
                value = expr.concrete_value.to_bytes(size, byteorder=self.byteorder)
            res = function(emu, addr, size, value)
            if res is not None:
                res_expr = claripy.BVV(res)
                if self.platform.byteorder == platforms.Byteorder.LITTLE:
                    # Fix byte order if needed.
                    # I don't know _why_ this is needed,
                    # but encoding the result as little-endian on a little-endian
                    # system produces the incorrect value in the machine state.
                    res_expr = claripy.Reverse(res_expr)
                return res_expr
            return res

        self.hook_memory_read_symbolic(start, end, sym_callback)

    def unhook_memory_read(self, start: int, end: int):
        if not self._initialized:
            self._read_hooks = list(
                filter(lambda x: x[0] != start and x[1] != end, self._read_hooks)
            )

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory hooking not supported once execution begins"
            )

        elif (start, end) not in self.state.scratch.mem_read_bps:
            raise exceptions.ConfigurationError(
                f"{hex(start)} - {hex(end)} is not hooked for reads"
            )
        else:
            bp = self.state.scratch.mem_read_bps[(start, end)]
            del self.state.scratch.mem_read_bps[(start, end)]
            self.state.inspect.remove_breakpoint("mem_read", bp=bp)

    def hook_memory_reads_symbolic(
        self,
        function: typing.Callable[
            [emulator.Emulator, int, int, claripy.ast.bv.BV],
            typing.Optional[claripy.ast.bv.BV],
        ],
    ) -> None:
        if not self._initialized:
            self._gb_read_hook = function

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory hooking not supported once execution begins"
            )

        elif self.state.scratch.global_read_bp is not None:
            raise exceptions.ConfigurationError(
                "Global memory read hook already present"
            )

        else:

            def read_callback(state):
                # the breakpoint action.
                addr = state.inspect.mem_read_address
                if not isinstance(addr, int):
                    addr = addr.concrete_value
                size = state.inspect.mem_read_length
                expr = state.inspect.mem_read_expr

                res = function(ConcreteAngrEmulator(state, self), addr, size, expr)

                if self.platform.byteorder == platforms.Byteorder.LITTLE:
                    # fix byte order if needed.
                    # i don't know _why_ this is needed,
                    # but encoding the result as little-endian on a little-endian
                    # system produces the incorrect value in the machine state.
                    res = claripy.Reverse(res)
                state.inspect.mem_read_expr = res

                # An update to angr means some operations on `state`
                # will clobber this variable, which causes bad problems.
                # Unclobber it, just in case
                state.inspect.action_attrs_set = True

            bp = self.state.inspect.b(
                "mem_read",
                when=angr.BP_AFTER,
                action=read_callback,
            )
            self.state.scratch.global_read_bp = bp

    def hook_memory_reads(
        self,
        function: typing.Callable[
            [emulator.Emulator, int, int, bytes], typing.Optional[bytes]
        ],
    ):
        # Symbolic version does most of the lifting;
        # we just need to translate the values
        def sym_callback(
            emu: emulator.Emulator, addr: int, size: int, expr: claripy.ast.bv.BV
        ) -> typing.Optional[claripy.ast.bv.BV]:
            if not isinstance(emu, ConcreteAngrEmulator):
                raise TypeError("Where'd you get your emu?")
            if expr.symbolic:
                # Have the solver handle binding resolution for us.
                try:
                    values = emu.state.solver.eval_atmost(expr, 1)
                    if len(values) < 1:
                        raise exceptions.AnalysisError(f"No possible values fpr {expr}")
                    value = values[0].to_bytes(size, byteorder=self.byteorder)
                    log.info("Collapsed symbolic {expr} to {values[0]:x} for MMIO")
                except angr.errors.SimUnsatError:
                    raise exceptions.AnalysisError(f"No possible values for {expr}")
                except angr.errors.SimValueError:
                    raise exceptions.AnalysisError(
                        f"Unbound value for MMIO write to {hex(addr)}: {expr}"
                    )
            else:
                value = expr.concrete_value.to_bytes(size, byteorder=self.byteorder)
            res = function(emu, addr, size, value)
            if res is not None:
                return claripy.BVV(res)
            return res

        self.hook_memory_reads_symbolic(sym_callback)

    def unhook_memory_reads(self) -> None:
        if not self._initialized:
            self._gb_read_hook = None

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory unhooking not supported once execution begins"
            )

        elif self.state.scratch.global_read_bp is not None:
            raise exceptions.ConfigurationError("Global memory read hook not present")

        else:
            bp = self.state.scratch.global_read_bp
            self.state.scratch.global_read_bp = None
            self.state.inspect.remove_breakpoint("mem_read", bp)

    def hook_memory_write_symbolic(
        self,
        start: int,
        end: int,
        function: typing.Callable[
            [emulator.Emulator, int, int, claripy.ast.bv.BV], None
        ],
    ) -> None:
        if not self._initialized:
            self._write_hooks.append((start, end, function))

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory hooking not supported once execution begins"
            )

        elif (start, end) in self.state.scratch.mem_write_bps:
            raise exceptions.ConfigurationError(
                f"{hex(start)} - {hex(end)} already hooked for writes"
            )

        else:

            def write_condition(state):
                write_start = state.inspect.mem_write_address
                if not isinstance(write_start, int):
                    if write_start.symbolic:
                        # Need to use the solver to resolve this.
                        try:
                            values = state.solver.eval_atmost(write_start, 1)
                            if len(values) < 1:
                                return False
                            write_start = values[0]
                        except angr.errors.SimUnsatError:
                            return False
                        except angr.errors.SimValueError:
                            return False
                    else:
                        write_start = write_start.concrete_value
                    # Populate concrete value back to the inspect struct
                    state.inspect.mem_write_address = write_start
                write_size = state.inspect.mem_write_length

                if write_size is None:
                    # Some ISAs don't populate mem_write_length.
                    # Infer length from value
                    write_size = len(state.inspect.mem_write_expr) // 8
                    # Populate concrete value back to the inspect struct
                    state.inspect.mem_write_length = write_size
                write_end = write_start + write_size

                rng = range(start, end)
                access_rng = range(write_start, write_end)
                return (
                    write_start in rng
                    or write_end - 1 in rng
                    or start in access_rng
                    or end - 1 in access_rng
                )

            def write_callback(state):
                addr = state.inspect.mem_write_address
                size = state.inspect.mem_write_length
                expr = state.inspect.mem_write_expr

                if size is None:
                    size = len(expr)

                function(ConcreteAngrEmulator(state, self), addr, size, expr)

                # An update to angr means some operations on `state`
                # will clobber this variable, which causes bad problems.
                # Unclobber it, just in case
                state.inspect.action_attrs_set = True

            bp = self.state.inspect.b(
                "mem_write",
                when=angr.BP_BEFORE,
                condition=write_condition,
                action=write_callback,
            )
            self.state.scratch.mem_write_bps[(start, end)] = bp

    def hook_memory_write(
        self,
        start: int,
        end: int,
        function: typing.Callable[[emulator.Emulator, int, int, bytes], None],
    ) -> None:
        def sym_callback(
            emu: emulator.Emulator, addr: int, size: int, expr: claripy.ast.bv.BV
        ) -> None:
            if not isinstance(emu, ConcreteAngrEmulator):
                raise TypeError("Where'd you get your emu?")
            if expr.symbolic:
                # Have the solver handle binding resolution for us.
                try:
                    values = emu.state.solver.eval_atmost(expr, 1)
                    if len(values) < 1:
                        raise exceptions.AnalysisError(f"No possible values fpr {expr}")
                    value = values[0].to_bytes(size, byteorder=self.byteorder)
                    log.info("Collapsed symbolic {expr} to {values[0]:x} for MMIO")
                except angr.errors.SimUnsatError:
                    raise exceptions.AnalysisError(f"No possible values for {expr}")
                except angr.errors.SimValueError:
                    raise exceptions.AnalysisError(
                        f"Unbound value for MMIO write to {hex(addr)}: {expr}"
                    )
            else:
                value = expr.concrete_value.to_bytes(size, byteorder=self.byteorder)
            function(emu, addr, size, value)

        self.hook_memory_write_symbolic(start, end, sym_callback)

    def unhook_memory_write(self, start: int, end: int) -> None:
        if not self._initialized:
            self._write_hooks = list(
                filter(lambda x: x[0] != start or x[1] != end, self._write_hooks)
            )

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory hooking not supported once execution begins"
            )

        elif (start, end) not in self.state.scratch.mem_write_bps:
            raise exceptions.ConfigurationError(
                f"{hex(start)} - {hex(end)} is not hooked for writes"
            )

        else:
            bp = self.state.scratch.mem_write_bps[(start, end)]
            del self.state.scratch.mem_write_bps[(start, end)]
            self.state.inspect.remove_breakpoint("mem_write", bp=bp)

    def hook_memory_writes_symbolic(
        self,
        function: typing.Callable[
            [emulator.Emulator, int, int, claripy.ast.bv.BV], None
        ],
    ) -> None:
        if not self._initialized:
            self._gb_write_hook = function

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory hooking not supported once execution begins"
            )

        elif self.state.scratch.global_write_bp is not None:
            raise exceptions.ConfigurationError(
                "Global memory write hook already present"
            )

        else:

            def write_callback(state):
                addr = state.inspect.mem_write_address
                if not isinstance(addr, int):
                    addr = addr.concrete_value
                size = state.inspect.mem_write_length
                expr = state.inspect.mem_write_expr
                function(ConcreteAngrEmulator(state, self), addr, size, expr)

                # An update to angr means some operations on `state`
                # will clobber this variable, which causes bad problems.
                # Unclobber it, just in case
                state.inspect.action_attrs_set = True

            bp = self.state.inspect.b(
                "mem_write",
                when=angr.BP_BEFORE,
                action=write_callback,
            )

            self.state.scratch.global_write_bp = bp

    def hook_memory_writes(
        self, function: typing.Callable[[emulator.Emulator, int, int, bytes], None]
    ) -> None:
        def sym_callback(
            emu: emulator.Emulator, addr: int, size: int, expr: claripy.ast.bv.BV
        ) -> None:
            if not isinstance(emu, ConcreteAngrEmulator):
                raise TypeError("Where'd you get your emu?")
            if expr.symbolic:
                # Have the solver handle binding resolution for us.
                try:
                    values = emu.state.solver.eval_atmost(expr, 1)
                    if len(values) < 1:
                        raise exceptions.AnalysisError(f"No possible values fpr {expr}")
                    value = values[0].to_bytes(size, byteorder=self.byteorder)
                    log.info("Collapsed symbolic {expr} to {values[0]:x} for MMIO")
                except angr.errors.SimUnsatError:
                    raise exceptions.AnalysisError(f"No possible values for {expr}")
                except angr.errors.SimValueError:
                    raise exceptions.AnalysisError(
                        f"Unbound value for MMIO write to {hex(addr)}: {expr}"
                    )
            else:
                value = expr.concrete_value.to_bytes(size, byteorder=self.byteorder)
            function(emu, addr, size, value)

        self.hook_memory_writes_symbolic(sym_callback)

    def unhook_memory_writes(self) -> None:
        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Memory unhooking not supported once execution begins"
            )
        if self.state.scratch.global_write_bp is not None:
            raise exceptions.ConfigurationError("Global memory write hook not present")

        bp = self.state.scratch.global_read_bp
        self.state.scratch.global_read_bp = None
        self.state.inspect.remove_breakpoint("mem_write", bp)

    def _step(self, single_insn: bool):
        """Common routine for all step functions.

        This is rather verbose, so let's only write it once.

        Arguments:
            single_insn: True to step one instruction.  False to step to the end of the block
        """
        # If we have not been initialized, initialize ourself
        if not self._initialized:
            self.initialize()

        # As soon as we start executing, disable value access
        self._dirty = True
        if self._linear:
            if self.state._ip.concrete_value not in self.state.scratch.func_bps:
                disas = self.state.block(opt_level=0).disassembly
                if disas is not None and len(disas.insns) > 0:
                    log.info(f"Stepping through {disas.insns[0]}")
                else:
                    # Capstone only supports a subset of the instructions supported by LibVEX.
                    # I can only disassemble what I can disassemble.
                    log.info(f"Stepping through {self.state._ip} (untranslatable!)")
            else:
                log.info(f"Stepping through {self.state._ip} (hook)")

        # Step execution once, however the user asked for it.
        if single_insn:
            # Not all architectures support single-step execution.
            # In particular, angr can't lift delay slot ISAs one instruction at a time,
            # since it has to lift the instruction and the slot as one unit.
            if not self.machdef.supports_single_step:
                raise exceptions.ConfigurationError(
                    f"AngrEmulator does not support single-instruction stepping for {self.platform}"
                )
            num_inst = 1
        else:
            num_inst = None
        self.mgr.step(
            num_inst=num_inst,
            successor_func=self.machdef.successors,
            thumb=self.machdef.is_thumb,
        )

        # Test for exceptional states
        if len(self.mgr.errored) > 0:
            raise exceptions.EmulationError(
                self.mgr.errored[0].error
            ) from self.mgr.errored[0].error

        # Handle linear execution mode
        if self._linear:
            if len(self.mgr.active) > 1:
                log.warn("Path diverged!  Detailes stored in simulation manager.")
                raise exceptions.EmulationStop("Path diverged in linear mode")
            elif len(self.mgr.active) > 0:
                self.state = self.mgr.active[0]
            elif len(self.mgr.deadended) > 0:
                self.state = self.mgr.deadended[0]
            elif len(self.mgr.unconstrained) > 0:
                self.state = self.mgr.unconstrained[0]
            else:
                raise exceptions.AnalysisError(
                    "No states in expected stashes for linear execution"
                )

        # Filter out exited or invalid states
        self.mgr.move(
            from_stash="active",
            to_stash="unconstrained",
            filter_func=lambda x: (x._ip.symbolic),
        )

        def filter_func(state):
            ip = state._ip.concrete_value
            if (
                not self.state.scratch.bounds.is_empty()
                and not self.state.scratch.bounds.contains_value(ip)
            ):
                return True
            if not self.state.scratch.memory_map.contains_value(ip):
                return True
            if ip in self.state.scratch.exit_points:
                return True
            return False

        self.mgr.move(
            from_stash="active", to_stash="deadended", filter_func=filter_func
        )

        # Stop if we're out of active states
        if len(self.mgr.active) == 0:
            raise exceptions.EmulationStop()

    def step_instruction(self) -> None:
        self._step(True)

    def step_block(self) -> None:
        self._step(False)

    def step(self) -> None:
        self._step(True)

    def run(self):
        log.info("Starting angr run")
        try:
            # Continue stepping as long as we have steps.
            while True:
                self._step(False)
        except exceptions.EmulationStop:
            return

    def visit_states(
        self,
        function: typing.Callable[[emulator.Emulator], None],
        stash: str = "active",
    ) -> None:
        """Visit every state in the selected frontier

        This lets you work around the fact that most operations
        only work at emulation start, or in linear mode.
        """
        if not self._initialized:
            raise NotImplementedError("Cannot visit states before initialization")
        for s in self.mgr.stashes[stash]:
            function(ConcreteAngrEmulator(s, self))

    def enable_linear(self):
        """Enable linear execution

        This doesn't actually concretize anything;
        it just kills execution when it hits an unconstrained branch.
        """
        if self._dirty:
            raise NotImplementedError(
                "Enabling linear mode not supported once execution begins"
            )
        self._linear = True
        log.warn("Linear execution mode enabled")

    def get_bounds(self) -> typing.List[typing.Tuple[int, int]]:
        if not self._initialized:
            return list(self._code_bounds)

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing bounds not supported once execution begins"
            )
        return list(self.state.scratch.bounds.ranges)

    def add_bound(self, start: int, end: int) -> None:
        if not self._initialized:
            self._code_bounds.append((start, end))

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing bounds not supported once execution begins"
            )

        else:
            self.state.scratch.bounds.add_range((start, end))

    def remove_bound(self, start: int, end: int) -> None:
        if not self._initialized:
            self._code_bounds = list(
                filter(lambda x: x[0] != start or x[1] != end, self._code_bounds)
            )

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing bounds not supported once execution begins"
            )

        else:
            self.state.scratch.bounds.remove_range((start, end))

    def get_exit_points(self) -> typing.Set[int]:
        if not self._initialized:
            return set(self._exit_points)

        if self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing exit points not supported once execution begins"
            )
        return set(self.state.scratch.exit_points)

    def add_exit_point(self, address: int) -> None:
        if not self._initialized:
            self._exit_points.add(address)
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing exit points not supported once execution begins"
            )
        else:
            self.state.scratch.exit_points.add(address)

    def remove_exit_point(self, address: int) -> None:
        if not self._initialized:
            self._exit_points.remove(address)

        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing exit points not supported once execution begins"
            )

        else:
            self.state.scratch.exit_points.remove(address)

    def add_constraint(self, expr: claripy.ast.bool.Bool) -> None:
        if not self._initialized:
            self._constraints.append(expr)
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing constraints not supported once execution begins"
            )
        else:
            self.state.solver.add(expr)

    def get_constraints(self) -> typing.List[claripy.ast.bool.Bool]:
        if not self._initialized:
            return list(self._constraints)
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing constraints not supported once execution begins"
            )
        else:
            return list(self.state.solver.constraints)

    def satisfiable(
        self,
        extra_constraints: typing.List[claripy.ast.bool.Bool] = [],
    ) -> bool:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Solver not available before initialization"
            )
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing solver not supported once execution begins"
            )
        return self.state.solver.satisfiable(extra_constraints=extra_constraints)

    def eval_atmost(self, expr: claripy.ast.bv.BV, most: int) -> typing.List[int]:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Solver not available before initialization"
            )
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing solver not supported once execution begins"
            )
        try:
            return self.state.solver.eval_atmost(expr, most)
        except angr.errors.SimUnsatError:
            raise exceptions.UnsatError("No solutions")
        except angr.errors.SimValueError:
            raise exceptions.SymbolicValueError("Too many solutions")

    def eval_atleast(self, expr: claripy.ast.bv.BV, least: int) -> typing.List[int]:
        if not self._initialized:
            raise exceptions.ConfigurationError(
                "Solver not available before initialization"
            )
        elif self._dirty and not self._linear:
            raise NotImplementedError(
                "Accessing solver not supported once execution begins"
            )
        try:
            return self.state.solver.eval_atleast(expr, least)
        except angr.errors.SimUnsatError:
            raise exceptions.UnsatError("No solutions")
        except angr.errors.SimValueError:
            raise exceptions.SymbolicValueError("Not enough solutions")

    def add_extension(self, name: str, ext: typing.Any) -> None:
        """Add extra data to the machine state.

        This allows users to tie data structures
        to the emulator's execution states,
        allowing path-specific data tracking.

        Only one extension with a given name can be tied
        to a given state.  Usually, extensions are assigned
        before emulation starts, but it's perfectly
        possible to assign them later, either in linear mode,
        or through the emulator stub provided to a hook.

        All extensions must be compatible with `copy.deepcopy()`.

        Arguments:
            name: ID string used to retrieve the extension
            ext: The extension

        Raises:
            KeyError: `name` is already taken
        """

        if not self._initialized:
            if name in self._extensions:
                raise KeyError(f"Extension with name {name} exists")
            self._extensions[name] = ext
        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot extend state once execution has begun")
        else:
            if name in self.state.scratch.extensions:
                raise KeyError(f"Extension with name {name} exists")
            self.state.scratch.extensions[name] = ext

    def get_extension(self, name: str) -> typing.Any:
        """Fetch extra data from the machine state

        Arguments:
            name: ID string used to retrieve the extension

        Returns:
            Whatever was assigned to `name` using `add_extension`

        Raises:
            KeyError: If no extension exists tied to `name`
        """
        if not self._initialized:
            return self._extensions[name]
        elif self._dirty and not self._linear:
            raise NotImplementedError("Cannot extend state once execution has begun")
        else:
            return self.state.scratch.extensions[name]

    @property
    def byteorder(self) -> typing.Literal["big", "little"]:
        """Get the byteorder string for this platform

        Returns:
            'big' or 'little', appropriately-typed as literals.
        """
        if self.platform.byteorder == platforms.Byteorder.BIG:
            return "big"
        elif self.platform.byteorder == platforms.Byteorder.LITTLE:
            return "little"
        else:
            raise ValueError("Unsupported byteorder!")

    def __repr__(self):
        if self._initialized:
            return f"Angr ({self.mgr})"
        else:
            return "Angr (Uninitialized)"


class ConcreteAngrEmulator(AngrEmulator):
    """
    "Emulator" for angr state access

    The primary AngrEmulator class is designed for symbolic execution;
    it represents many states being explored in parallel.
    As such, the notion of accessing state doesn't work; which state do you want?

    For hook evaluation, we do need an emulator-like object
    that can access a single state.  Hence, this wrapper class.

    NOTE: This is NOT a full emulator!
    Think of it more as a view onto an AngrEmulator instance.
    If you want to explore a single path using angr,
    use AngrEmulator and call enable_linear() before exploration.
    """

    @property
    def PAGE_SIZE(self):
        # Page Size should be inherited from the parent.
        return self.pagesize

    def __init__(self, state: angr.SimState, parent: AngrEmulator):
        # Do NOT call the superclass constructor.
        # It initializes the angr project, and we've already done that.
        self._initialized: bool = True
        self._dirty: bool = False
        self._linear: bool = False

        # Force-initialize relevant state from the parent
        self.platform: platforms.Platform = parent.platform
        self.proj: angr.Project = parent.proj
        self.state: angr.SimState = state
        self.platdef: platforms.PlatformDef = parent.platdef
        self.machdef: AngrMachineDef = parent.machdef
        self.pagesize: int = parent.PAGE_SIZE

    # Function hooking is not supported;
    # it relies on state global to the angr project, not individual states.
    def hook_function(
        self, address: int, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        raise NotImplementedError("Function hooking not supported inside a hook.")

    def unhook_function(self, address: int) -> None:
        raise NotImplementedError("Function hooking not supported inside a hook.")

    # Execution is not supported; this is not a complete emulator.
    def run(self) -> None:
        raise NotImplementedError("Running not supported inside a hook.")

    def _step(self, single_insn: bool) -> None:
        raise NotImplementedError("Stepping not supported inside a hook.")

    def __repr__(self):
        return f"Angr Hook ({self.state})"


__all__ = ["AngrEmulator"]



import logging
import re
import struct
import typing

from ....emulators import Emulator
from ....platforms import Byteorder
from ..cstd import ArgumentType, CStdModel, VariadicContext
from .utils import _emu_strlen

log = logging.getLogger(__name__)

# Conversion specifiers:
#
# Group 1: Zero or more flags
#
# Group 2: Zero or one field width specifiers
# - '[0-9]+': Explicit width
# - '*': Next argument is an int storing the width
#
# Group 3: Zero or one precision specifiers
# - '.': Zero precision
# - '.[0-9]+': Explicit precision
# - '.*': Next argument is an int storing the width
# - NOTE: Default precision for floats is six.
#
# Group 4: Unused; part of the precision pattern
#
# Group 5: Zero or one length specifiers
#
# Group 6: Conversion ID

# Signed decimal int conversion
#
# Allowed flags:
# - '0': Justify the field with zeros
# - ' ': Add a space before a positive number
# - '-': Left-justify the field.  Default is right-justify
# - '+': Always include a sign symbol
#
# Allowed lengths:
# - 'hh': char
# - 'h': short
# - '': int
# - 'l': long
# - 'll', 'q': long long
# - 'z', 'Z': ssize_t
# - 'j': intmax_t
# - 't': ptrdiff_t
#
# Allowed conversions:
# - d, i: signed decimal integer
sdecint_re = re.compile(
    "%([0 \\-+]*)([0-9]*|[*])((\\.[0-9]*|\\.[*])?)(hh|h|l|ll|q|z|Z|j|t|)(d|i)"
)

# Unsigned decimal int conversion
#
# Allowed flags:
# - '0': Justify the field with zeros
# - '-': Left-justify the field.  Default is right-justify
#
# Allowed lengths:
# - 'hh': char
# - 'h': short
# - '': int
# - 'l': long
# - 'll', 'q': long long
# - 'z', 'Z': size_t
# - 'j': intmax_t
# - 't': ptrdiff_t
#
# Allowed conversions:
# - u: unsigned decimal integer
udecint_re = re.compile(
    "%([0\\-]*)([0-9]*|[*])((\\.[0-9]*|\\.[*])?)(hh|h|l|ll|q|z|Z|j|t|)(u)"
)

# Unsigned non-decimal int conversion
#
# Allowed flags:
# - '#': Add '0o' or '0x' prefix, as appropriate
# - '0': Justify the field with zeros
# - '-': Left-justify the field.  Default is right-justify
#
# Allowed lengths:
# - 'hh': char
# - 'h': short
# - '': int
# - 'l': long
# - 'll', 'q': long long
# - 'z', 'Z': size_t
# - 'j': intmax_t
# - 't': ptrdiff_t
#
# Allowed conversions:
# - o: unsigned octal integer
# - x: unsigned hexadecimal integer, lowercase
# - X: unsigned hexadecimal integer, uppercase
uint_re = re.compile(
    "%([#0\\-]*)([0-9]*|[*])((\\.[0-9]*|\\.[*])?)(hh|h|l|ll|q|z|Z|j|t|)(o|x|X)"
)

# Scientific notation conversion
#
# D.DDeDD
#
# Allowed flags:
# - '#': Include decimal and fractional component if zero
# - '0': Justify the field with zeros
# - ' ': Add a space before a positive number
# - '-': Left-justify the field.  Default is right-justify
# - '+': Always include a sign symbol
#
# Allowed lengths
# - '': double (the compiler better damned well have gotten this right).
# - 'L': long double
#
# Allowed conversions:
# - 'e': Scientific notation, lowercase
# - 'E': Scientific notation, uppercase
sci_re = re.compile("%([#0 \\-+]*)([0-9]*|[*])((\\.[0-9]*|\\.[*])?)(L|)(e|E)")

# Floating point conversion
#
# DD.DD
#
# Allowed flags:
# - '#': Include decimal and fractional component if zero
# - '0': Justify the field with zeros
# - ' ': Add a space before a positive number
# - '-': Left-justify the field.  Default is right-justify
# - '+': Always include a sign symbol
#
# Allowed lengths
# - '': double (the compiler better damned well have gotten this right).
# - 'L': long double
#
# Allowed conversions:
# - 'e': Floating-point, lowercase
# - 'E': Floating-point, uppercase
float_re = re.compile("%([#0 \\-+]*)([0-9]*|[*])((\\.[0-9]*|\\.[*])?)(L|)(f|F)")

# Size-sensitive floating point conversion
#
# DD.DD or D.DDeDD, depending on magnitude
#
# Allowed flags:
# - '#': Include decimal and fractional component if zero
# - '0': Justify the field with zeros
# - ' ': Add a space before a positive number
# - '-': Left-justify the field.  Default is right-justify
# - '+': Always include a sign symbol
#
# Allowed lengths
# - '': double (the compiler better damned well have gotten this right).
# - 'L': long double
#
# Allowed conversions:
# - 'g': Floating-point, lowercase
# - 'G': Floating-point, uppercase
scifloat_re = re.compile("%([#0 \\-+]*)([0-9]*|[*])((\\.[0-9]*|\\.[*])?)(L|)(g|G)")

# Hexadecimal floating point conversion
#
# 0xXX.XX
#
# Allowed flags:
# - '#': Include decimal and fractional component if zero
# - '0': Justify the field with zeros
# - ' ': Add a space before a positive number
# - '-': Left-justify the field.  Default is right-justify
# - '+': Always include a sign symbol
#
# Allowed lengths
# - '': double (the compiler better damned well have gotten this right).
# - 'L': long double
#
# Allowed conversions:
# - 'a': Hexadecimal floating-point, lowercase
# - 'A': Hexadecimal floating-point, uppercase
hexfloat_re = re.compile("%([#0 \\-+]*)([0-9]*|[*])((\\.[0-9]+|\\.[*])?)(L|)(g|G)")

# Character conversion
#
# c
#
# Allowed flags:
# - ' ': Add a space before a string
# - '-': Left-justify the field.  Default is right-justify
#
# Allowed Lengths:
# - ' ': unsigned char
# - 'l': wchar_t
#
# Allowed conversions:
# - 'c': Character
char_re = re.compile("%([ \\-]*)([0-9]*|[*])()()(l|)(c)")

# String conversion
#
# cccccccc
#
# Allowed flags:
# - ' ': Add a space before a string
# - '-': Left-justify the field.  Default is right-justify
#
# Allowed lengths
# - '': char *
# - 'l': wchar_t *
#
# Allowed conversions:
# - s: String
str_re = re.compile("%([ \\-]*)([0-9]*|[*])((\\.[0-9]+|\\.[*])?)(l|)(s)")

# Pointer conversion
#
# 0xXXXXXXXX
#
# Allowed flags:
# - '-': Left-justify the field.
#
# Allowed lengths:
# - '': void *
#
# Allowed conversions:
# - p: pointer
pointer_re = re.compile("%([\\-]?)([0-9]*|[*])()()()(p)")

# Length of current output
#
# <prints nothing>
#
# Allowed flags:
# - None
#
# Allowed lengths:
# - 'hh': char *
# - 'h': short *
# - '': int *
# - 'l': long *
# - 'll', 'q': long long *
# - 'z', 'Z': size_t *
# - 'j': intmax_t *
# - 't': ptrdiff_t *
#
# Allowed conversions:
# - 'n': Fetch length of current output
len_re = re.compile("%()()()()(hh|h|l|ll|q|z|Z|j|t|)(n)")

# Strerror conversion
#
# ssssssss # result of strerror(errno)
#
# Allowed flags:
# - ' ': Add a space before a string
# - '-': Left-justify the field.  Default is right-justify
#
# Allowed lengths:
# - '': Default; this doesn't take an argument.
#
# Allowed conversions:
# - varargs: VariadicContext, m: strerror
strerror_re = re.compile("%([ \\-]*)([0-9]*|[*])()()()(m)")

# Percent sign
#
# %
#
# Allowed flags:
# - None
#
# Allowed lengths:
# - None
#
# Allowed conversions:
# - '%': Print a literal '%'
percent_re = re.compile("%%")


class FormatConversionError(Exception):
    pass


def handle_flags(flags: str) -> str:
    out = ""
    if "-" in flags:
        out += "<"
    else:
        out += ">"

    if "+" in flags:
        out += "+"

    if " " in flags:
        out += " "

    if "0" in flags:
        out += "0"

    if "#" in flags:
        out += "#"

    return out


def handle_width(width: str, varargs: VariadicContext, emulator: Emulator) -> str:
    if width == "*":
        width_int = varargs.get_next_argument(ArgumentType.INT, emulator)
        assert isinstance(width_int, int)
        return str(width_int)
    else:
        return width


def handle_precision(
    precision: str, varargs: VariadicContext, emulator: Emulator, use_default=True
) -> str:
    if precision == ".*":
        precision_int = varargs.get_next_argument(ArgumentType.INT, emulator)
        assert isinstance(precision_int, int)
        return "." + str(precision_int)
    elif precision == ".":
        return ".0"
    elif precision == "":
        if use_default:
            return ".6"
        else:
            return ""
    else:
        return precision


def handle_double_arg(
    length: str, varargs: VariadicContext, emulator: Emulator
) -> float:
    if length == "":
        val = varargs.get_next_argument(ArgumentType.DOUBLE, emulator)
    elif length == "L":
        raise NotImplementedError("Type 'long double' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")
    assert isinstance(val, float)
    return val


def handle_sdecint(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    # Conversion is irrelevant; they're aliases for the same thing.

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle precision
    # Precision in ints acts like a second width,
    # which always has the '0' flag enabled.
    # This isn't implemented in python, which means we need to do it piecewise
    if precision != "":
        raise NotImplementedError("TODO: Implemente precision for ints")

    # Handle length
    if length in ("hh", "h", ""):
        # Variadics don't go smaller than 'int'
        val = varargs.get_next_argument(ArgumentType.INT, emulator)
    elif length == "l":
        val = varargs.get_next_argument(ArgumentType.LONG, emulator)
    elif length in ("ll", "q"):
        val = varargs.get_next_argument(ArgumentType.LONGLONG, emulator)
    elif length in ("z", "Z"):
        val = varargs.get_next_argument(ArgumentType.SSIZE_T, emulator)
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    assert isinstance(val, int)

    fmt += "d}"

    return fmt.format(val)


def handle_udecint(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    # Conversion is irrelevant; they're aliases for the same thing.

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle precision
    # Precision in ints acts like a second width,
    # which always has the '0' flag enabled.
    # This isn't implemented in python, which means we need to do it piecewise
    if precision != "":
        raise NotImplementedError("TODO: Implemente precision for ints")

    # Handle length
    if length in ("hh", "h", ""):
        # Variadics don't go smaller than 'int'
        val = varargs.get_next_argument(ArgumentType.UINT, emulator)
    elif length == "l":
        val = varargs.get_next_argument(ArgumentType.ULONG, emulator)
    elif length in ("ll", "q"):
        val = varargs.get_next_argument(ArgumentType.ULONGLONG, emulator)
    elif length in ("z", "Z"):
        val = varargs.get_next_argument(ArgumentType.SIZE_T, emulator)
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")
    assert isinstance(val, int)

    fmt += "d}"

    return fmt.format(val)


def handle_uint(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    conv = m.group(6)

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    # In C, octals can have both '#' and '0' flags.
    # In practice, the '#' is redundant, and forbidden by Python
    if "0" in flags:
        flags = flags.replace("#", "")

    # Handle width
    width = handle_width(width, varargs, emulator)

    # Handle precision
    # Precision in ints acts like a second width,
    # which always has the '0' flag enabled.
    # This isn't implemented in python, which means we need to do it piecewise
    if precision != "":
        raise NotImplementedError("TODO: Implemente precision for ints")

    # Handle length
    if length in ("hh", "h", ""):
        # Variadics don't go smaller than 'int'
        val = varargs.get_next_argument(ArgumentType.UINT, emulator)
    elif length == "l":
        val = varargs.get_next_argument(ArgumentType.ULONG, emulator)
    elif length in ("ll", "q"):
        val = varargs.get_next_argument(ArgumentType.ULONGLONG, emulator)
    elif length in ("z", "Z"):
        val = varargs.get_next_argument(ArgumentType.SIZE_T, emulator)
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    assert isinstance(val, int)

    fmt = "{:"
    fmt += flags
    fmt += width
    fmt += conv
    fmt += "}"
    res = fmt.format(val)
    if "#" in flags and conv == "o":
        if width != "":
            int_width = int(width)
        else:
            int_width = 0
        extend = len(res) == int_width
        res = res.replace("0o", "0")
        if extend:
            res = " " + res

    return res


def handle_sci(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    conv = m.group(6)

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle precision
    precision = handle_precision(precision, varargs, emulator)
    fmt += precision

    # Handle length
    val = handle_double_arg(length, varargs, emulator)

    fmt += conv
    fmt += "}"

    out = fmt.format(val)

    # Handle negative NaN.
    # Python prints it as positive NaN.
    byteval = struct.pack("<d", val)
    intval = int.from_bytes(byteval, "little")
    if intval == 0xFFF8000000000000:
        text = ("{:" + conv + "}").format(val)
        index = out.index(text)
        if index == 0:
            out = out.replace(text, "-" + text)
        else:
            out = out[: index - 1] + "-" + text + out[index + 3 :]

    return out


def handle_float(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    conv = m.group(6)

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle precision
    precision = handle_precision(precision, varargs, emulator)
    fmt += precision

    # Handle length
    val = handle_double_arg(length, varargs, emulator)

    fmt += conv
    fmt += "}"

    out = fmt.format(val)

    # Handle negative NaN.
    # Python prints it as positive NaN.
    byteval = struct.pack("<d", val)
    intval = int.from_bytes(byteval, "little")
    if intval == 0xFFF8000000000000:
        text = ("{:" + conv + "}").format(val)
        index = out.index(text)
        if index == 0:
            out = out.replace(text, "-" + text)
        else:
            out = out[: index - 1] + "-" + text + out[index + 3 :]

    return out


def handle_scifloat(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    conv = m.group(6)

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle precision
    precision = handle_precision(precision, varargs, emulator)
    fmt += precision

    # Handle length
    val = handle_double_arg(length, varargs, emulator)

    fmt += conv
    fmt += "}"

    out = fmt.format(val)

    # Handle negative NaN.
    # Python prints it as positive NaN.
    byteval = struct.pack("<d", val)
    intval = int.from_bytes(byteval, "little")

    if intval == 0xFFF8000000000000:
        text = ("{:" + conv + "}").format(val)
        index = out.index(text)
        if index == 0:
            out = out.replace(text, "-" + text)
        else:
            out = out[: index - 1] + "-" + text + out[index + 3 :]

    return out


def handle_hexfloat(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    # Python doesn't have an easy conversion for hex floats
    # And I am not writing one unless you absolutely need it.
    raise NotImplementedError("Hexadecimal float conversion not supported")


def handle_char(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    length = m.group(5)
    conv = m.group(6)

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle length
    if length == "":
        val = varargs.get_next_argument(ArgumentType.UINT, emulator)
    elif length == "l":
        raise NotImplementedError("Type 'wchar_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    assert isinstance(val, int)
    val = val & 0xFF

    fmt += conv
    fmt += "}"

    return fmt.format(val)


def handle_str(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)
    precision = m.group(3)
    length = m.group(5)
    conv = m.group(6)

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    # Handle precision
    # Highly fortunately, this works the same in python
    precision = handle_precision(precision, varargs, emulator, use_default=False)
    fmt += precision

    # Handle length
    if length == "":
        addr = varargs.get_next_argument(ArgumentType.POINTER, emulator)
    elif length == "l":
        raise NotImplementedError("Type 'wchar_t *' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    assert isinstance(addr, int)

    val_len = _emu_strlen(emulator, addr)
    val_bytes = emulator.read_memory(addr, val_len)
    val = val_bytes.decode("utf-8")

    fmt += conv
    fmt += "}"

    return fmt.format(val)


def handle_pointer(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    flags = m.group(1)
    width = m.group(2)

    fmt = "{:"

    # Handle flags
    # These mean the same for python as they do for C,
    # with the exception of '-' in C being '<' in Python
    flags = handle_flags(flags)
    fmt += flags + "#"

    # Handle width
    width = handle_width(width, varargs, emulator)
    fmt += width

    val = varargs.get_next_argument(ArgumentType.POINTER, emulator)
    assert isinstance(val, int)

    fmt += "x}"

    return fmt.format(val)


def handle_len(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    length = m.group(5)

    addr = varargs.get_next_argument(ArgumentType.POINTER, emulator)
    assert isinstance(addr, int)

    val = len(output)
    if length == "hh":
        val = val & 0xFF
        width = 1
    elif length == "h":
        val = val & 0xFFFF
        width = 2
    elif length == "":
        val = val & varargs._int_inv_mask
        width = 4
    elif length in ("l", "z", "Z"):
        val = val & varargs._long_inv_mask
        width = 4 if ArgumentType.LONG in varargs._four_byte_types else 8
    elif length in ("ll", "q"):
        val = val & varargs._long_long_inv_mask
        width = 8
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    if varargs.platform.byteorder == Byteorder.LITTLE:
        data = val.to_bytes(width, "little")
    else:
        data = val.to_bytes(width, "big")

    emulator.write_memory(addr, data)

    return ""


def handle_strerror(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    # This is weird.  It's a glibc extension, but gcc errors if you use it.
    # If you actually see this, please open a bug ticket.
    raise NotImplementedError(
        "%m conversion (strerror) not implemented; if you see this, submit a ticket."
    )


def handle_percent(
    output: str, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> str:
    # If this has any configurable options, I will be most displeased.
    return "%"


def parse_printf_format(model: CStdModel, fmt: str, emulator: Emulator) -> str:
    """Create a string based on a printf format

    This doesn't model all features of the printf family.
    The following features are missing:

    - The %a and %A hexadecimal float conversions
    - The %m strerror conversion
    - intmax_t integer size
    - ptrdiff_t integer size
    - long double float size

    Arguments:
        model: The parent model
        fmt: The format string
        emulator: The emulator to query

    Returns:
        A string formatted according to 'fmt'
    """
    handlers: typing.List[
        typing.Tuple[
            re.Pattern, typing.Callable[[str, VariadicContext, re.Match, Emulator], str]
        ]
    ] = [
        (sdecint_re, handle_sdecint),
        (udecint_re, handle_udecint),
        (uint_re, handle_uint),
        (sci_re, handle_sci),
        (float_re, handle_float),
        (scifloat_re, handle_scifloat),
        (hexfloat_re, handle_hexfloat),
        (char_re, handle_char),
        (str_re, handle_str),
        (pointer_re, handle_pointer),
        (len_re, handle_len),
        (strerror_re, handle_strerror),
        (percent_re, handle_percent),
    ]

    varargs = model.get_varargs()

    orig_fmt = fmt
    output = ""

    while len(fmt) > 0:
        if fmt.startswith("%"):
            matched = False
            for regex, handler in handlers:
                if isinstance(regex, str):
                    raise Exception(f"String: {regex}")
                m = regex.match(fmt)
                if m is not None:
                    try:
                        output += handler(output, varargs, m, emulator)
                    except FormatConversionError as e:
                        log.error(f"Bad format conversion: {e.args[0]}")
                        log.error(orig_fmt)
                        log.error(" " * (len(orig_fmt) - len(fmt)) + "^")
                        raise e
                    except Exception as e:
                        log.error(f"Exception processing conversion: {type(e)}: {e}")
                        log.error(orig_fmt)
                        log.error(" " * (len(orig_fmt) - len(fmt)) + "^")
                        raise e

                    fmt = fmt[len(m.group(0)) :]
                    matched = True
                    break

            if not matched:
                log.error("Bad format conversion: Unmatched conversion")
                log.error(orig_fmt)
                log.error(" " * (len(orig_fmt) - len(fmt)) + "^")
                raise Exception("Bad format conversion")

        else:
            output += fmt[0]
            fmt = fmt[1:]

    return output


__all__ = ["parse_printf_format"]



from ..... import emulators, platforms
from ...cstd import ArgumentType, CStdModel


class I386SysVModel(CStdModel):
    """Base class for C models using the AMD64 System V ABI"""

    platform = platforms.Platform(
        platforms.Architecture.X86_32, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.SYSTEMV

    _int_sign_mask = 0x80000000
    _int_inv_mask = 0xFFFFFFFF
    _long_sign_mask = 0x80000000
    _long_inv_mask = 0xFFFFFFFF
    _long_long_sign_mask = 0x8000000000000000
    _long_long_inv_mask = 0xFFFFFFFFFFFFFFFF

    _four_byte_types = {
        ArgumentType.INT,
        ArgumentType.UINT,
        ArgumentType.LONG,
        ArgumentType.ULONG,
        ArgumentType.SIZE_T,
        ArgumentType.SSIZE_T,
        ArgumentType.POINTER,
    }

    _eight_byte_types = {
        ArgumentType.LONGLONG,
        ArgumentType.ULONGLONG,
    }

    _four_byte_arg_regs = []
    _eight_byte_arg_regs = []

    _soft_float = False
    _variadic_soft_float = False
    _floats_are_doubles = False
    _float_arg_regs = []
    _double_arg_regs = []

    _init_stack_offset = 4
    _align_stack = False
    _eight_byte_reg_size = 1
    _double_reg_size = 1
    _four_byte_stack_size = 4
    _eight_byte_stack_size = 8
    _float_stack_size = 4
    _double_stack_size = 8

    def _return_4_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return a four-byte type"""
        emulator.write_register("eax", val)

    def _return_8_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return an eight-byte type"""
        hi = (val >> 32) & self._int_inv_mask
        lo = val & self._int_inv_mask

        emulator.write_register("eax", lo)
        emulator.write_register("edx", hi)

    def _return_float(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a float"""
        raise NotImplementedError("i386 System-V uses x87 registers to return floats")

    def _return_double(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a double"""
        raise NotImplementedError("i386 System-V uses x87 registers to return doubles")



import os
import time
import typing

from ....emulators import Emulator
from ..cstd import ArgumentType, CStdModel
from .utils import _emu_strlen

# NOTE: time_t and clock_t are longs.
# As such, all your 32-bit routers are vulnerable to the 2038 bug.


class TimeModel(CStdModel):
    def time_struct_to_tuple(self, ptr: int, emulator: Emulator):
        # Load the fields out of the struct
        tm_sec = self.read_integer(ptr, ArgumentType.INT, emulator)
        tm_min = self.read_integer(ptr + 0x4, ArgumentType.INT, emulator)
        tm_hour = self.read_integer(ptr + 0x8, ArgumentType.INT, emulator)
        tm_mday = self.read_integer(ptr + 0xC, ArgumentType.INT, emulator)
        tm_mon = self.read_integer(ptr + 0x10, ArgumentType.INT, emulator)
        tm_year = self.read_integer(ptr + 0x14, ArgumentType.INT, emulator)
        tm_wday = self.read_integer(ptr + 0x18, ArgumentType.INT, emulator)
        tm_yday = self.read_integer(ptr + 0x1C, ArgumentType.INT, emulator)
        tm_isdst = self.read_integer(ptr + 0x20, ArgumentType.INT, emulator)

        # Load data into a python struct
        # A few of the fields are represented a bit differently
        out = (
            tm_year + 1900,
            tm_mon + 1,
            tm_mday,
            tm_hour,
            tm_min,
            tm_sec,
            tm_wday - 1,
            tm_yday + 1,
            tm_isdst,
        )

        return out

    def tuple_to_time_struct(
        self,
        tp: typing.Tuple[int, int, int, int, int, int, int, int, int],
        ptr: int,
        emulator: Emulator,
    ):
        (
            tm_year,
            tm_mon,
            tm_mday,
            tm_hour,
            tm_min,
            tm_sec,
            tm_wday,
            tm_yday,
            tm_isdst,
        ) = tp

        # Convert the Python representation to C's representation
        tm_year -= 1900
        tm_mon -= 1
        tm_wday += 1
        tm_yday -= 1

        # Store the fields back to the struct
        self.write_integer(ptr, tm_sec, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x4, tm_min, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x8, tm_hour, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0xC, tm_mday, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x10, tm_mon, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x14, tm_year, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x18, tm_wday, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x1C, tm_yday, ArgumentType.INT, emulator)
        self.write_integer(ptr + 0x20, tm_isdst, ArgumentType.INT, emulator)


class Time(TimeModel):
    name = "time"

    # time_t time(time_t *tloc);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.LONG

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        intval = int(time.time())

        if ptr != 0:
            self.write_integer(ptr, intval, ArgumentType.LONG, emulator)

        self.set_return_value(emulator, intval)


class Localtime(TimeModel):
    name = "localtime"

    # struct tm *localtime(const time_t *timep);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    # Need a static struct tm
    # NOTE: This looks bigger, because GNU struct tms have extra fields.
    static_space_required = 64

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        tv = self.read_integer(ptr, ArgumentType.LONG, emulator)

        timetuple = time.localtime(tv)

        assert self.static_buffer_address is not None

        self.tuple_to_time_struct(timetuple, self.static_buffer_address, emulator)
        self.set_return_value(emulator, self.static_buffer_address)


class Gmtime(TimeModel):
    name = "gmtime"

    # struct tm *gmtime(const time_t *timep);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    # Need a static struct tm
    # NOTE: This looks bigger, because GNU struct tms have extra fields.
    static_space_required = 64

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        tv = self.read_integer(ptr, ArgumentType.LONG, emulator)

        timetuple = time.gmtime(tv)

        assert self.static_buffer_address is not None

        self.tuple_to_time_struct(timetuple, self.static_buffer_address, emulator)
        self.set_return_value(emulator, self.static_buffer_address)


class Ctime(TimeModel):
    name = "ctime"

    # struct tm *ctime(const time_t *timep);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    static_space_required = 26

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        tv = self.read_integer(ptr, ArgumentType.LONG, emulator)

        timestr = time.ctime(tv)
        timebytes = timestr.encode("utf-8") + b"\n\0"

        assert len(timebytes) <= 26
        assert self.static_buffer_address is not None

        emulator.write_memory(self.static_buffer_address, timebytes)
        self.set_return_value(emulator, self.static_buffer_address)


class Asctime(TimeModel):
    name = "asctime"

    # char *asctime(const struct tm *tp);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    # Need a 26-byte static buffer to store the output string
    static_space_required = 26

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        timetuple = self.time_struct_to_tuple(ptr, emulator)

        # FIXME: mktime is timezone dependent
        # This is actually something of a pain to handle.  For now, assume UTC.
        old_tz = None
        if "TZ" in os.environ:
            old_tz = os.environ["TZ"]
        os.environ["TZ"] = "UTC"
        time.tzset()

        print(timetuple)
        timestr = time.asctime(timetuple)
        print(timestr)

        if old_tz is None:
            del os.environ["TZ"]
        else:
            os.environ["TZ"] = old_tz
        time.tzset()

        timebytes = timestr.encode("utf-8") + b"\n\0"

        assert len(timebytes) <= 26
        assert self.static_buffer_address is not None

        emulator.write_memory(self.static_buffer_address, timebytes)

        self.set_return_value(emulator, self.static_buffer_address)


class Strftime(TimeModel):
    name = "strftime"

    # size_t strftime(char *dst, size_t max, const char *fmt, const struct tm *tp);
    argument_types = [
        ArgumentType.POINTER,
        ArgumentType.SIZE_T,
        ArgumentType.POINTER,
        ArgumentType.POINTER,
    ]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        dst = self.get_arg1(emulator)
        max = self.get_arg2(emulator)
        fmt = self.get_arg3(emulator)
        ptr = self.get_arg4(emulator)

        assert isinstance(dst, int)
        assert isinstance(max, int)
        assert isinstance(fmt, int)
        assert isinstance(ptr, int)

        fmtlen = _emu_strlen(emulator, fmt)
        fmtbytes = emulator.read_memory(fmt, fmtlen)
        fmtstr = fmtbytes.decode("utf-8")

        timetuple = self.time_struct_to_tuple(ptr, emulator)
        timestr = time.strftime(fmtstr, timetuple)
        timebytes = timestr.encode("utf-8")
        timebytes += b"\0"

        if len(timebytes) > max:
            self.set_return_value(emulator, 0)
        else:
            emulator.write_memory(dst, timebytes)
            self.set_return_value(emulator, len(timebytes))


class Difftime(TimeModel):
    name = "difftime"

    # double difftime(time_t time0, time_t time1);
    argument_types = [ArgumentType.LONG, ArgumentType.LONG]
    return_type = ArgumentType.DOUBLE

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        time0 = self.get_arg1(emulator)
        time1 = self.get_arg2(emulator)

        assert isinstance(time0, int)
        assert isinstance(time1, int)

        # Times are longs.  Why does this return a double?
        # The world is full of mysteries...
        diff = float(time1 - time0)

        self.set_return_value(emulator, diff)


class Mktime(TimeModel):
    name = "mktime"

    # time_t mktime(struct tm *tp);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.LONG

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        timetuple = self.time_struct_to_tuple(ptr, emulator)
        timefloat = time.mktime(timetuple)

        self.set_return_value(emulator, int(timefloat))


class Clock(TimeModel):
    name = "clock"

    # clock_t clock(void)
    argument_types = []
    return_type = ArgumentType.LONG

    def model(self, emulator: Emulator) -> None:
        super().model(emulator)
        floatval = time.clock_gettime(time.CLOCK_PROCESS_CPUTIME_ID)
        floatval /= time.clock_getres(time.CLOCK_PROCESS_CPUTIME_ID)

        self.set_return_value(emulator, int(floatval))


__all__ = [
    "Time",
    "Localtime",
    "Gmtime",
    "Ctime",
    "Asctime",
    "Strftime",
    "Difftime",
    "Mktime",
    "Clock",
]



from ... import platforms
from .. import state
from . import cpu


class PowerPC(cpu.CPU):
    """CPU state for 32-bit PowerPC."""

    def __init__(self, wordsize):
        super().__init__()
        # *** General Purpose Registers ***
        # NOTE: Used expressive names for GPRs and FPRs.
        # gasm just refers to GPRs and FPRS by number.
        # They use the same numbers; it's very annoying.
        self.r0 = state.Register("r0", size=wordsize)
        self.add(self.r0)
        # NOTE: GPR 1 is also the stack pointer.
        self.r1 = state.Register("r1", size=wordsize)
        self.add(self.r1)
        self.sp = state.RegisterAlias("sp", self.r1, size=wordsize, offset=0)
        self.add(self.sp)
        self.r2 = state.Register("r2", size=wordsize)
        self.add(self.r2)
        self.r3 = state.Register("r3", size=wordsize)
        self.add(self.r3)
        self.r4 = state.Register("r4", size=wordsize)
        self.add(self.r4)
        self.r5 = state.Register("r5", size=wordsize)
        self.add(self.r5)
        self.r6 = state.Register("r6", size=wordsize)
        self.add(self.r6)
        self.r7 = state.Register("r7", size=wordsize)
        self.add(self.r7)
        self.r8 = state.Register("r8", size=wordsize)
        self.add(self.r8)
        self.r9 = state.Register("r9", size=wordsize)
        self.add(self.r9)
        self.r10 = state.Register("r10", size=wordsize)
        self.add(self.r10)
        self.r11 = state.Register("r11", size=wordsize)
        self.add(self.r11)
        self.r12 = state.Register("r12", size=wordsize)
        self.add(self.r12)
        self.r13 = state.Register("r13", size=wordsize)
        self.add(self.r13)
        self.r14 = state.Register("r14", size=wordsize)
        self.add(self.r14)
        self.r15 = state.Register("r15", size=wordsize)
        self.add(self.r15)
        self.r16 = state.Register("r16", size=wordsize)
        self.add(self.r16)
        self.r17 = state.Register("r17", size=wordsize)
        self.add(self.r17)
        self.r18 = state.Register("r18", size=wordsize)
        self.add(self.r18)
        self.r19 = state.Register("r19", size=wordsize)
        self.add(self.r19)
        self.r20 = state.Register("r20", size=wordsize)
        self.add(self.r20)
        self.r21 = state.Register("r21", size=wordsize)
        self.add(self.r21)
        self.r22 = state.Register("r22", size=wordsize)
        self.add(self.r22)
        self.r23 = state.Register("r23", size=wordsize)
        self.add(self.r23)
        self.r24 = state.Register("r24", size=wordsize)
        self.add(self.r24)
        self.r25 = state.Register("r25", size=wordsize)
        self.add(self.r25)
        self.r26 = state.Register("r26", size=wordsize)
        self.add(self.r26)
        self.r27 = state.Register("r27", size=wordsize)
        self.add(self.r27)
        self.r28 = state.Register("r28", size=wordsize)
        self.add(self.r28)
        self.r29 = state.Register("r29", size=wordsize)
        self.add(self.r29)
        self.r30 = state.Register("r30", size=wordsize)
        self.add(self.r30)
        # NOTE: GPR 31 is also the base pointer
        self.r31 = state.Register("r31", size=wordsize)
        self.add(self.r31)
        self.bp = state.RegisterAlias("bp", self.r31, size=wordsize, offset=0)
        self.add(self.bp)

        # Floating Point Registers
        # Always 8 bytes, regardless of wordsize.
        self.f0 = state.Register("f0", size=8)
        self.add(self.f0)
        self.f1 = state.Register("f1", size=8)
        self.add(self.f1)
        self.f2 = state.Register("f2", size=8)
        self.add(self.f2)
        self.f3 = state.Register("f3", size=8)
        self.add(self.f3)
        self.f4 = state.Register("f4", size=8)
        self.add(self.f4)
        self.f5 = state.Register("f5", size=8)
        self.add(self.f5)
        self.f6 = state.Register("f6", size=8)
        self.add(self.f6)
        self.f7 = state.Register("f7", size=8)
        self.add(self.f7)
        self.f8 = state.Register("f8", size=8)
        self.add(self.f8)
        self.f9 = state.Register("f9", size=8)
        self.add(self.f9)
        self.f10 = state.Register("f10", size=8)
        self.add(self.f10)
        self.f11 = state.Register("f11", size=8)
        self.add(self.f11)
        self.f12 = state.Register("f12", size=8)
        self.add(self.f12)
        self.f13 = state.Register("f13", size=8)
        self.add(self.f13)
        self.f14 = state.Register("f14", size=8)
        self.add(self.f14)
        self.f15 = state.Register("f15", size=8)
        self.add(self.f15)
        self.f16 = state.Register("f16", size=8)
        self.add(self.f16)
        self.f17 = state.Register("f17", size=8)
        self.add(self.f17)
        self.f18 = state.Register("f18", size=8)
        self.add(self.f18)
        self.f19 = state.Register("f19", size=8)
        self.add(self.f19)
        self.f20 = state.Register("f20", size=8)
        self.add(self.f20)
        self.f21 = state.Register("f21", size=8)
        self.add(self.f21)
        self.f22 = state.Register("f22", size=8)
        self.add(self.f22)
        self.f23 = state.Register("f23", size=8)
        self.add(self.f23)
        self.f24 = state.Register("f24", size=8)
        self.add(self.f24)
        self.f25 = state.Register("f25", size=8)
        self.add(self.f25)
        self.f26 = state.Register("f26", size=8)
        self.add(self.f26)
        self.f27 = state.Register("f27", size=8)
        self.add(self.f27)
        self.f28 = state.Register("f28", size=8)
        self.add(self.f28)
        self.f29 = state.Register("f29", size=8)
        self.add(self.f29)
        self.f30 = state.Register("f30", size=8)
        self.add(self.f30)
        self.f31 = state.Register("f31", size=8)
        self.add(self.f31)

        # *** Pointer Registers ***
        # Program Counter.
        # Not really a register; nothing can access it directly
        self.pc = state.Register("pc", size=wordsize)
        self.add(self.pc)

        # Link Register
        self.lr = state.Register("lr", size=wordsize)
        self.add(self.lr)

        # Counter Register
        # Acts either as a loop index, or a branch target register
        # Only `ctr` and `lr` can act as branch targets.
        self.ctr = state.Register("ctr", size=wordsize)
        self.add(self.ctr)

        # *** Condition Registers ***
        # Condition Register
        # The actual condition register `cr` is a single 32-bit register,
        # but it's broken into eight 4-bit fields which are accessed separately.
        self.cr0 = state.Register("cr0", size=1)  # Integer condition bits
        self.add(self.cr0)
        self.cr1 = state.Register("cr1", size=1)  # Floatibg point condition bits
        self.add(self.cr1)
        self.cr2 = state.Register("cr2", size=1)
        self.add(self.cr2)
        self.cr3 = state.Register("cr3", size=1)
        self.add(self.cr3)
        self.cr4 = state.Register("cr4", size=1)
        self.add(self.cr4)
        self.cr5 = state.Register("cr5", size=1)
        self.add(self.cr5)
        self.cr6 = state.Register("cr6", size=1)
        self.add(self.cr6)
        self.cr7 = state.Register("cr7", size=1)
        self.add(self.cr7)

        # Integer Exception Register
        self.xer = state.Register("xer", size=4)
        self.add(self.xer)

        # Floating Point Status and Control Register
        self.fpscr = state.Register("fpscr", size=4)
        self.add(self.fpscr)

        # TODO: This only focuses on the user-facing registrers.
        # ppc has a huge number of privileged registers.
        # Extend this as needed.


class PowerPC32(PowerPC):
    """CPU state for 32-bit PowerPC."""

    platform = platforms.Platform(
        platforms.Architecture.POWERPC32, platforms.Byteorder.BIG
    )

    def __init__(self):
        super().__init__(4)


class PowerPC64(PowerPC):
    """CPU state for 64-bit PowerPC."""

    platform = platforms.Platform(
        platforms.Architecture.POWERPC64, platforms.Byteorder.BIG
    )

    def __init__(self):
        super().__init__(8)



from ....c99.signal import Raise, Signal
from ..systemv import I386SysVModel


class I386SysVRaise(Raise, I386SysVModel):
    pass


class I386SysVSignal(Signal, I386SysVModel):
    pass


__all__ = [
    "I386SysVRaise",
    "I386SysVSignal",
]



from ....platforms import Architecture, Byteorder
from .machdef import PandaMachineDef


class AMD64MachineDef(PandaMachineDef):
    arch = Architecture.X86_64
    byteorder = Byteorder.LITTLE

    panda_arch = "x86_64"

    # I'm going to define all the ones we are making possible as of now
    # I need to submit a PR to change to X86 32 bit and to include eflags
    _registers_64 = {
        "rax",
        "rbx",
        "rcx",
        "rdx",
        "rsi",
        "rdi",
        "rsp",
        "rbp",
        "rip",
        "r8",
        "r9",
        "r10",
        "r11",
        "r12",
        "r13",
        "r14",
        "r15",
    }
    _registers_general = {
        "eax",
        "ebx",
        "ecx",
        "edx",
        "esi",
        "edi",
        "esp",
        "ebp",
        "eip",
        "r8d",
        "r9d",
        "r10d",
        "r11d",
        "r12d",
        "r13d",
        "r14d",
        "r15d",
    }
    _registers_short = {
        "ax",
        "bx",
        "cx",
        "dx",
        "si",
        "di",
        "sp",
        "bp",
        "r8w",
        "r9w",
        "r10w",
        "r11w",
        "r12w",
        "r13w",
        "r14w",
        "r15w",
    }
    _registers_byte = {
        "al",
        "bl",
        "cl",
        "dl",
        "r8b",
        "r9b",
        "r10b",
        "r11b",
        "r12b",
        "r13b",
        "r14b",
        "r15b",
        "ah",
        "bh",
        "ch",
        "dh",
    }
    _registers_flags = {"rflags", "eflags", "flags"}
    _registers_seg = {"es", "cs", "ss", "ds", "fs", "gs"}
    _registers_control = {"cr0", "cr1", "cr2", "cr3", "cr4"}
    _registers_debug = {f"dr{i}" for i in range(0, 16)} - {"dr4", "dr5"}
    _registers_mmr = {"gdtr": "gdt", "idtr": "idt", "tr": "tr", "ldtr": "ldt"}
    _registers_x87 = {
        "fpr0",
        "fpr1",
        "fpr2",
        "fpr3",
        "fpr4",
        "fpr5",
        "fpr6",
        "fpr7",
        "fctrl",
        "fstat",
        "ftag",
        "fip",
        "fdp",
        "fop",
    }
    _registers_mmx = {"mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7"}
    _registers_xmm = {
        "xmm0",
        "xmm1",
        "xmm2",
        "xmm3",
        "xmm4",
        "xmm5",
        "xmm6",
        "xmm7",
        "xmm8",
        "xmm9",
        "xmm10",
        "xmm11",
        "xmm12",
        "xmm13",
        "xmm14",
        "xmm15",
    }
    _registers_ymm = {
        "ymm0",
        "ymm1",
        "ymm2",
        "ymm3",
        "ymm4",
        "ymm5",
        "ymm6",
        "ymm7",
        "ymm8",
        "ymm9",
        "ymm10",
        "ymm11",
        "ymm12",
        "ymm13",
        "ymm14",
        "ymm15",
    }
    _registers_pc = {"pc": "rip", "eip": "eip", "ip": None}
    _registers_absent = {"dil", "sil", "spl", "bpl", "cr8"}

    _registers = {}
    _registers = _registers | {i: i for i in _registers_64}
    _registers = _registers | {i: i for i in _registers_general}
    _registers = _registers | {i: i for i in _registers_short}
    _registers = _registers | {i: i for i in _registers_byte}
    _registers = _registers | {i: None for i in _registers_flags}
    _registers = _registers | {i: i for i in _registers_seg}
    _registers = _registers | {i: i for i in _registers_control}
    _registers = _registers | {i: None for i in _registers_debug}
    _registers = _registers | {i: j for i, j in _registers_mmr.items()}
    _registers = _registers | {i: None for i in _registers_x87}
    _registers = _registers | {i: None for i in _registers_mmx}
    _registers = _registers | {i: i for i in _registers_xmm}
    _registers = _registers | {i: None for i in _registers_ymm}
    _registers = _registers | {i: j for i, j in _registers_pc.items()}
    _registers = _registers | {i: None for i in _registers_absent}



from .... import emulators
from ..c99.utils import _emu_strlen
from ..cstd import ArgumentType, CStdModel


class Basename(CStdModel):
    name = "basename"

    # NOTE: There are two versions of basename().
    #
    # POSIX and GNU have competing specifications.
    # They have the following differences:
    #
    # - GNU always uses a static buffer for its returns
    # - GNU returns "" in case of trailing separators
    # - GNU returns "" in case of the root path

    # char *basename(char *path);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    # Use the GNU memory semantics
    static_space_required = 0x1000

    def __init__(self, address: int):
        super().__init__(address)

        # Override this to change how you parse paths
        self.separator = b"/"

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        pathptr = self.get_arg1(emulator)

        assert isinstance(pathptr, int)

        # Not actually a loop; I just want to use break.
        while True:
            if pathptr == 0:
                # Case: pathptr is NULL; return a static buffer containing '.'
                path = b"."
                break

            pathlen = _emu_strlen(emulator, pathptr)
            path = emulator.read_memory(pathptr, pathlen)

            if path == self.separator:
                # Case: Path is the root path: Return itself
                break

            if path[-1] == self.separator[0]:
                # Case: Path ends in separator: Delete the trailing separator
                path = path[0:-1]

            if self.separator not in path:
                # Case: no path separator: Return copy of path
                break

            # Case: Path has a separator in it: Return the trailing substring of path
            idx = path.rindex(self.separator)
            path = path[idx + 1 :]
            break

        assert self.static_buffer_address is not None

        emulator.write_memory(self.static_buffer_address, path + b"\0")
        self.set_return_value(emulator, self.static_buffer_address)


class Dirname(CStdModel):
    name = "dirname"

    # char *dirname(char *path);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def __init__(self, address: int):
        super().__init__(address)

        # Override this to change how you parse paths
        self.separator = b"/"

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        pathptr = self.get_arg1(emulator)

        assert isinstance(pathptr, int)

        if pathptr == 0:
            # Case: pathptr is NULL; return a static buffer containing '.'
            raise NotImplementedError("Requires a static buffer")

        pathlen = _emu_strlen(emulator, pathptr)
        path = emulator.read_memory(pathptr, pathlen)

        if path == self.separator:
            # Case: Path is '/': Return itself
            print("Root case")
            self.set_return_value(emulator, pathptr)
            return

        if self.separator not in path:
            # Case: No path separator; return '.'
            emulator.write_memory(pathptr, b".\0")
            self.set_return_value(emulator, pathptr)
            return

        if path[-1] == self.separator[0]:
            path = path[:-2]

        idx = path.rindex(self.separator)
        if idx == 0:
            # Case: Path's lowest '/' is the root path; return '/'
            emulator.write_memory(pathptr + idx + 1, b"\0")
        else:
            # Case: Path has a '/' in it.  Replace it with \0 and return original pointer
            emulator.write_memory(pathptr + idx, b"\0")
        self.set_return_value(emulator, pathptr)


__all__ = ["Basename", "Dirname"]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import PowerPCSysVModel


class PowerPCSysVAbs(Abs, PowerPCSysVModel):
    pass


class PowerPCSysVLAbs(LAbs, PowerPCSysVModel):
    pass


class PowerPCSysVLLAbs(LLAbs, PowerPCSysVModel):
    pass


class PowerPCSysVAtof(Atof, PowerPCSysVModel):
    pass


class PowerPCSysVAtoi(Atoi, PowerPCSysVModel):
    pass


class PowerPCSysVAtol(Atol, PowerPCSysVModel):
    pass


class PowerPCSysVAtoll(Atoll, PowerPCSysVModel):
    pass


class PowerPCSysVCalloc(Calloc, PowerPCSysVModel):
    pass


class PowerPCSysVDiv(Div, PowerPCSysVModel):
    pass


class PowerPCSysVLDiv(LDiv, PowerPCSysVModel):
    pass


class PowerPCSysVLLDiv(LLDiv, PowerPCSysVModel):
    pass


class PowerPCSysVExit(Exit, PowerPCSysVModel):
    pass


class PowerPCSysVFree(Free, PowerPCSysVModel):
    pass


class PowerPCSysVMalloc(Malloc, PowerPCSysVModel):
    pass


class PowerPCSysVQSort(QSort, PowerPCSysVModel):
    pass


class PowerPCSysVRand(Rand, PowerPCSysVModel):
    pass


class PowerPCSysVRealloc(Realloc, PowerPCSysVModel):
    pass


class PowerPCSysVSrand(Srand, PowerPCSysVModel):
    pass


class PowerPCSysVAbort(Abort, PowerPCSysVModel):
    pass


class PowerPCSysVAtexit(Atexit, PowerPCSysVModel):
    pass


class PowerPCSysVGetenv(Getenv, PowerPCSysVModel):
    pass


class PowerPCSysVMblen(Mblen, PowerPCSysVModel):
    pass


class PowerPCSysVMbstowcs(Mbstowcs, PowerPCSysVModel):
    pass


class PowerPCSysVMbtowc(Mbtowc, PowerPCSysVModel):
    pass


class PowerPCSysVSystem(System, PowerPCSysVModel):
    pass


class PowerPCSysVWcstombs(Wcstombs, PowerPCSysVModel):
    pass


class PowerPCSysVWctomb(Wctomb, PowerPCSysVModel):
    pass


__all__ = [
    "PowerPCSysVAbs",
    "PowerPCSysVLAbs",
    "PowerPCSysVLLAbs",
    "PowerPCSysVAtof",
    "PowerPCSysVAtoi",
    "PowerPCSysVAtol",
    "PowerPCSysVAtoll",
    "PowerPCSysVCalloc",
    "PowerPCSysVDiv",
    "PowerPCSysVLDiv",
    "PowerPCSysVLLDiv",
    "PowerPCSysVExit",
    "PowerPCSysVFree",
    "PowerPCSysVMalloc",
    "PowerPCSysVQSort",
    "PowerPCSysVRand",
    "PowerPCSysVRealloc",
    "PowerPCSysVSrand",
    "PowerPCSysVAbort",
    "PowerPCSysVAtexit",
    "PowerPCSysVGetenv",
    "PowerPCSysVMblen",
    "PowerPCSysVMbstowcs",
    "PowerPCSysVMbtowc",
    "PowerPCSysVSystem",
    "PowerPCSysVWcstombs",
    "PowerPCSysVWctomb",
]



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import AArch64SysVModel


class AArch64SysVBsdSignal(BsdSignal, AArch64SysVModel):
    pass


class AArch64SysVKill(Kill, AArch64SysVModel):
    pass


class AArch64SysVKillpg(Killpg, AArch64SysVModel):
    pass


class AArch64SysVPthreadKill(PthreadKill, AArch64SysVModel):
    pass


class AArch64SysVPthreadSigmask(PthreadSigmask, AArch64SysVModel):
    pass


class AArch64SysVSigaction(Sigaction, AArch64SysVModel):
    pass


class AArch64SysVSigaddset(Sigaddset, AArch64SysVModel):
    pass


class AArch64SysVSigaltstack(Sigaltstack, AArch64SysVModel):
    pass


class AArch64SysVSigdelset(Sigdelset, AArch64SysVModel):
    pass


class AArch64SysVSigemptyset(Sigemptyset, AArch64SysVModel):
    pass


class AArch64SysVSigfillset(Sigfillset, AArch64SysVModel):
    pass


class AArch64SysVSighold(Sighold, AArch64SysVModel):
    pass


class AArch64SysVSigignore(Sigignore, AArch64SysVModel):
    pass


class AArch64SysVSiginterrupt(Siginterrupt, AArch64SysVModel):
    pass


class AArch64SysVSigismember(Sigismember, AArch64SysVModel):
    pass


class AArch64SysVSigpause(Sigpause, AArch64SysVModel):
    pass


class AArch64SysVSigpending(Sigpending, AArch64SysVModel):
    pass


class AArch64SysVSigprocmask(Sigprocmask, AArch64SysVModel):
    pass


class AArch64SysVSigqueue(Sigqueue, AArch64SysVModel):
    pass


class AArch64SysVSigrelse(Sigrelse, AArch64SysVModel):
    pass


class AArch64SysVSigset(Sigset, AArch64SysVModel):
    pass


class AArch64SysVSigsuspend(Sigsuspend, AArch64SysVModel):
    pass


class AArch64SysVSigtimedwait(Sigtimedwait, AArch64SysVModel):
    pass


class AArch64SysVSigwait(Sigwait, AArch64SysVModel):
    pass


class AArch64SysVSigwaitinfo(Sigwaitinfo, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVBsdSignal",
    "AArch64SysVKill",
    "AArch64SysVKillpg",
    "AArch64SysVPthreadKill",
    "AArch64SysVPthreadSigmask",
    "AArch64SysVSigaction",
    "AArch64SysVSigaddset",
    "AArch64SysVSigaltstack",
    "AArch64SysVSigdelset",
    "AArch64SysVSigemptyset",
    "AArch64SysVSigfillset",
    "AArch64SysVSighold",
    "AArch64SysVSigignore",
    "AArch64SysVSiginterrupt",
    "AArch64SysVSigismember",
    "AArch64SysVSigpause",
    "AArch64SysVSigpending",
    "AArch64SysVSigprocmask",
    "AArch64SysVSigqueue",
    "AArch64SysVSigrelse",
    "AArch64SysVSigset",
    "AArch64SysVSigsuspend",
    "AArch64SysVSigtimedwait",
    "AArch64SysVSigwait",
    "AArch64SysVSigwaitinfo",
]



import abc
import typing

from .... import platforms, utils
from ... import state
from .. import memory


class Stack(memory.Memory):
    """A stack-like region of memory with convenient operations like push and pop."""

    def __init__(self, address: int, size: int) -> None:
        self.stack_pointer = address + size
        super().__init__(address, size)

    @property
    @abc.abstractmethod
    def platform(self) -> platforms.Platform:
        """The platform for which this stack is intended."""
        pass

    @classmethod
    def get_platform(cls) -> platforms.Platform:
        """Retrieve the platform for this stack."""
        if not isinstance(cls.platform, platforms.Platform):
            raise TypeError(f"{cls.__name__}.platform is not a Platform object")
        return cls.platform

    def get_pointer(self) -> int:
        """Get the current stack pointer.

        Returns:
          The current value of the stack pointer.
        """
        return self.stack_pointer

    @abc.abstractmethod
    def get_alignment(self) -> int:
        """Get the alignment for this stack.

        Returns:
          The alignment for this stack.
        """
        pass

    @abc.abstractmethod
    def push(self, value: state.Value) -> int:
        """Push a value to the stack.

        Arguments:
            value: The value to be pushed.
        Returns:
            The stack pointer after the push.
        """
        pass

    def push_integer(self, integer: int, size: int, label: str) -> int:
        """Push an integer to the stack.

        Arguments:
            integer: The integer value to be pushed.
            size: The size in bytes for the integer on the stack.
            label: The label for the integer.
        Returns:
            The stack pointer after the push.
        """

        value = state.IntegerValue(integer, size, label)
        return self.push(value)

    def push_bytes(self, content: typing.Union[bytes, bytearray], label: str) -> int:
        """Push some bytes to the stack.

        Arguments:
            content: The bytes to push.
            label: The label for the bytes.

        Returns:
            The stack pointer after the push.
        """
        value = state.BytesValue(content, label)
        return self.push(value)

    def push_ctype(self, content: typing.Any, label: str) -> int:
        """Push some structured bytes to the stack.

        Arguments:
            content: The ctypes structured bytes.
            label: The label for the bytes.
        Returns:
            The stack pointer after the push.
        """
        value = state.Value.from_ctypes(content, label)
        return self.push(value)

    @classmethod
    def for_platform(cls, platform: platforms.Platform, address: int, size: int):
        """Create a stack for this platform.

        Arguments:
            platform: The platform for which this stack is intended.
            address: Start address for this stack.
            size: Size of requested stack, in bytes.
        """

        def check(x: typing.Type[Stack]):
            if x.get_platform():
                return (
                    x.get_platform().architecture == platform.architecture
                    and x.get_platform().byteorder == platform.byteorder
                )
            return False

        try:
            return utils.find_subclass(cls, check, address, size)
        except ValueError:
            raise ValueError(f"No stack for {platform}")


class DescendingStack(Stack):
    """A stack that grows down, i.e., a push will decrease the stack pointer."""

    def push(self, value: state.Value) -> int:
        self._is_safe(value)
        offset = self.stack_pointer - self.address - value.get_size()
        self[offset] = value
        self.stack_pointer -= value.get_size()
        return offset


__all__ = ["Stack"]



from ....c99.signal import Raise, Signal
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVRaise(Raise, LoongArch64SysVModel):
    pass


class LoongArch64SysVSignal(Signal, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVRaise",
    "LoongArch64SysVSignal",
]



import typing

from .... import platforms
from . import stack


class ARM32Stack(stack.DescendingStack):
    """A stack for an ARM 32-bit CPU"""

    def get_alignment(self) -> int:
        return 8

    @classmethod
    def initialize_stack(cls, argv: typing.List[bytes], *args, **kwargs):
        raise NotImplementedError("Stack initialization not implemented for ARM32")


class ARMv5tStack(ARM32Stack):
    """A stack for an ARMv5t 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V5T, platforms.Byteorder.LITTLE
    )


class ARMv6mStack(ARM32Stack):
    """A stack for an ARMv6m 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V6M, platforms.Byteorder.LITTLE
    )


class ARMv6mThumbStack(ARM32Stack):
    """A stack for an ARMv6mThumb 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V6M_THUMB, platforms.Byteorder.LITTLE
    )


class ARMv7mStack(ARM32Stack):
    """A stack for an ARMv7m 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V7M, platforms.Byteorder.LITTLE
    )


class ARMv7rStack(ARM32Stack):
    """A stack for an ARMv7r 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V7R, platforms.Byteorder.LITTLE
    )


class ARMv7aStack(ARM32Stack):
    """A stack for an ARMv7a 32-bit CPU"""

    platform = platforms.Platform(
        platforms.Architecture.ARM_V7A, platforms.Byteorder.LITTLE
    )



import typing
from dataclasses import dataclass

import jsons
import networkx as nx

from .. import hinting
from ..analyses.trace_execution_types import TraceElement, TraceRes


@dataclass(frozen=True)
class PointerHint(hinting.Hint):
    """We found a pointer

    Arguments:
        instruction: The instruction containing the pointer.
        pointer: The pointer.
    """

    instruction: typing.Any
    pointer: typing.Any


@dataclass(frozen=True)
class ControlFlowHint(hinting.Hint):
    """Represents control flow going from the from_instruction to the to_instruction.

    Arguments:
        from_instruction: The from instruction
        to_instruction: The to instruction
    """

    from_instruction: typing.Any
    to_instruction: typing.Any


@dataclass(frozen=True)
class CoverageHint(hinting.Hint):
    """Holds the a map of program counter to hit counter for an execution.

    Arguments:
        coverage: A map from program counter to hit count
    """

    coverage: typing.Dict[int, int]


@dataclass(frozen=True)
class ReachableCodeHint(hinting.Hint):
    """Indicates that we can get to a given program counter with symbolic execution.

    Arguments:
        address: The address we can reach
    """

    address: int


@dataclass(frozen=True)
class EmulationException(hinting.Hint):
    """Something went wrong emulating this instruction"""

    # instruction: typing.Any
    pc: int
    instruction_num: int
    exception: str


@dataclass(frozen=True)
class UnderSpecifiedValueHint(hinting.Hint):
    """Super class for UnderSpecified Value Hints"""

    instruction: typing.Any


@dataclass(frozen=True)
class UnderSpecifiedRegisterHint(UnderSpecifiedValueHint):
    """Represents a register whose value can't be fully determined from the environment.

    Arguments:
        register: The register in question
    """

    register: str


@dataclass(frozen=True)
class UnderSpecifiedMemoryHint(UnderSpecifiedValueHint):
    """Represents a memory range whose value can't be fully determined from the environment.

    Arguments:
        address: The address of the beginning of the range
        size: The size of the range
    """

    address: int
    size: int


@dataclass(frozen=True)
class UnderSpecifiedMemoryRefHint(UnderSpecifiedValueHint):
    """Represents a memory range whose value can't be fully determined from the environment.

    Arguments:
        address: The address of the beginning of the range
        size: The size of the range
    """

    base: typing.Tuple[str, int]
    index: typing.Tuple[str, int]
    offset: int


@dataclass(frozen=True)
class UnderSpecifiedAddressHint(UnderSpecifiedValueHint):
    """Represents a symbolic address that can't be resolved from the environment.
    Arguments:
        symbol: Name of the symbolic value
        addr:   Address expression containing the symbol
    """

    symbol: str
    addr: str


@dataclass(frozen=True)
class TypedUnderSpecifiedRegisterHint(UnderSpecifiedRegisterHint):
    typedef: str
    value: str


@dataclass(frozen=True)
class UntypedUnderSpecifiedRegisterHint(UnderSpecifiedRegisterHint):
    value: str


@dataclass(frozen=True)
class TypedUnderSpecifiedMemoryHint(UnderSpecifiedMemoryHint):
    typedef: str
    value: str


@dataclass(frozen=True)
class UntypedUnderSpecifiedMemoryHint(UnderSpecifiedMemoryHint):
    value: str


@dataclass(frozen=True)
class TypedUnderSpecifiedAddressHint(UnderSpecifiedAddressHint):
    typedef: str
    value: str


@dataclass(frozen=True)
class UntypedUnderSpecifiedAddressHint(UnderSpecifiedAddressHint):
    value: str


@dataclass(frozen=True)
class UnderSpecifiedBranchHint(UnderSpecifiedValueHint):
    """Represents a program fork based on an under-specified condition."""


@dataclass(frozen=True)
class UnderSpecifiedMemoryBranchHint(UnderSpecifiedBranchHint):
    """Represents conditional data flow with an under-specified conditional

    Arguments:
      addr: Offending address expression
      options: Possible evaluations of addr, paired with their guard expressions.
    """

    address: str
    options: typing.List[typing.Tuple[str, str]]


@dataclass(frozen=True)
class UnderSpecifiedControlBranchHint(UnderSpecifiedBranchHint):
    """Represents conditional control flow with an under-specified conditional

    Arguments:
      targets: Possible branch target addresses, paired with guard expressions.
    """

    targets: typing.Dict[str, str]


@dataclass(frozen=True)
class InputUseHint(UnderSpecifiedValueHint):
    """Represents an instruction at which some register input value is used,
       i.e. an information flow from input to some instruction

    Arguments:
      input_reg: The name of the register input value (source)
      instr: The instruction in which the input is used
      pc: program counter of that instruction
      use_reg: The name of the register in instr that is using the input value
    """

    input_register: str
    instruction_num: int
    use_register: str


@dataclass(frozen=True)
class TypeHint(hinting.Hint):
    """Super class for Type Hints"""

    pass


@dataclass(frozen=True)
class RegisterPointerHint(TypeHint):
    """Signal that a register is probably a pointer.

    Arguments:
        register: The register in question
    """

    register: str


@dataclass(frozen=True)
class RegisterPointsToHint(RegisterPointerHint):
    """Signal that a register is probably a pointer and points to a type.

    Arguments:
        type: The type in question
    """

    type: str


@dataclass(frozen=True)
class MemoryPointerHint(TypeHint):
    """Signal that a memory address is probably a pointer.

    Arguments:
       address: The address in question
    """

    address: int


@dataclass(frozen=True)
class MemoryPointsToHint(RegisterPointerHint):
    """Signal that a memory address is probably a pointer and points to a type.

    Arguments:
        type: The type in question
    """

    type: str


@dataclass(frozen=True)
class StructureHint(TypeHint):
    """Signals the probable layout of a struct

    Arguments:
        layout: A dictionary of offset to type
    """

    layout: typing.Dict[int, str]


@dataclass(frozen=True)
class OutputHint(hinting.Hint):
    registers: typing.Dict[str, str]
    memory: typing.Dict[int, str]


# This is used by the TraceExecutor


@dataclass(frozen=True)
class TraceExecutionHint(TypeHint):
    trace: typing.List[TraceElement]
    trace_digest: str
    seed: int
    emu_result: TraceRes
    exception: typing.Optional[Exception]
    exception_class: str

    # note this is just for logging purposes
    def to_json(self):
        tracel = [te.to_json() for te in self.trace]

        return {
            "trace": tracel,  # jsons.dump(self.trace),
            "trace_digest": self.trace_digest,
            "seed": jsons.dump(self.seed),
            "emu_result": jsons.dump(self.emu_result),
            "exception": jsons.dump(self.exception),
            "exception_class": str(type(self.exception)),
        }


# These next three are used by the colorizer


@dataclass(frozen=True)
class MemoryUnavailableHint(hinting.Hint):
    """Represents a load or store that was unavailable memory.

    Arguments:
      is_read: true if a load else a store
      size: size of read/write in bytes
      base_reg_name: name of base register (if known)
      base_reg_val: value of base register (if known)
      index_reg_name: name of index register (if known)
      index_reg_val: value of index register (if known)
      offset: offset (if known, else 0)
      scale: scale (if known, else 0)
      address: memory address of this value
      instruction: a smallworld instruction
      pc: program counter of that instruction
      instruction_num: for micro-execution the instr count
    """

    is_read: bool
    size: int
    base_reg_name: str
    base_reg_val: int
    index_reg_name: str
    index_reg_val: int
    offset: int
    scale: int
    address: int
    pc: int
    instruction_num: int


@dataclass(frozen=True)
class MemoryUnavailableSummaryHint(hinting.Hint):
    is_read: bool
    size: int
    base_reg_name: str
    index_reg_name: str
    offset: int
    scale: int
    pc: int
    count: int


@dataclass(frozen=True)
class DynamicValueHint(hinting.Hint):
    """Represents a concrete value either in a register or memory
    encountered during emulation-base analysis

    Arguments:
      instruction: a smallworld instruction
      pc: program counter of that instruction
      instruction_num: for micro-execution the instr count
      dynamic_value: this is the actual value
      size: the size of the value in bytes
      use: True if its a "use" of this value, else its a "def"
      new: True if its a new value, first sighting
    """

    # instruction: typing.Any
    pc: int
    instruction_num: int
    exec_id: int
    dynamic_value: int
    color: int
    size: int
    use: bool
    new: bool


@dataclass(frozen=True)
class DynamicRegisterValueHint(DynamicValueHint):
    """Represents a concrete register value encountered during
    analysis, either used or defined by some instruction.

    Arguments:
      reg_name: name of the register
      dynamic_value: this is the actual value as bytes
      use: True if its a "use" of this value, else its a "def"
      capstone_instruction: the instruction in capstone parlance
      pc: program counter of that instruction
      instruction_num: for micro-execution the instr count
      info: extra info about use or def if available
    """

    reg_name: str


@dataclass(frozen=True)
class DynamicMemoryValueHint(DynamicValueHint):
    """Represents a concrete memory value encountered during
    analysis, either used or defined by some instruction.

    Arguments:
      address: memory address of this value
      base: base address (if known, else 0)
      index: index (if known, else 0)
      scale: scale (if known, else 0)
      offset: offset (if known, else 0)
      dynamic_value: this is the actual value as bytes
      use: True if its a "use" of this value, else its a "def"
      capstone_instruction: the instruction in capstone parlance
      pc: program counter of that instruction
      instruction_num: for micro-execution the instr count
      info: extra info about use or def if available
    """

    address: int
    base: str
    index: str
    scale: int
    offset: int


@dataclass(frozen=True)
class DynamicValueSummaryHint(hinting.Hint):
    # instruction: typing.Any
    pc: int
    color: int
    size: int
    use: bool
    new: bool
    count: int
    num_micro_executions: int


@dataclass(frozen=True)
class DynamicMemoryValueSummaryHint(DynamicValueSummaryHint):
    base: str
    index: str
    scale: int
    offset: int


@dataclass(frozen=True)
class DynamicRegisterValueSummaryHint(DynamicValueSummaryHint):
    reg_name: str


@dataclass(frozen=True)
class DefUseGraphHint(hinting.Hint):
    graph: nx.MultiDiGraph


@dataclass(frozen=True)
class BranchesHint(hinting.Hint):
    branches: typing.List[int]


@dataclass(frozen=True)
class CoverageFrontierHint(hinting.Hint):
    coverage_frontier: typing.List[int]


__all__ = [
    "DynamicRegisterValueHint",
    "DynamicMemoryValueHint",
    "MemoryUnavailableHint",
    "DynamicRegisterValueSummaryHint",
    "DynamicMemoryValueSummaryHint",
    "MemoryUnavailableSummaryHint",
    "EmulationException",
    "CoverageHint",
    "ControlFlowHint",
    "ReachableCodeHint",
    "DefUseGraphHint",
    "TraceExecutionHint",
    "BranchesHint",
    "CoverageFrontierHint",
]



import struct

from ..... import emulators, platforms
from ...cstd import ArgumentType, CStdModel


class PowerPCSysVModel(CStdModel):
    """Base class for C models using the PowerPC System-V ABI"""

    platform = platforms.Platform(
        platforms.Architecture.POWERPC32, platforms.Byteorder.BIG
    )
    abi = platforms.ABI.SYSTEMV

    _int_sign_mask = 0x80000000
    _int_inv_mask = 0xFFFFFFFF
    _long_sign_mask = 0x80000000
    _long_inv_mask = 0xFFFFFFFF
    _long_long_sign_mask = 0x8000000000000000
    _long_long_inv_mask = 0xFFFFFFFFFFFFFFFF

    _four_byte_types = {
        ArgumentType.INT,
        ArgumentType.UINT,
        ArgumentType.LONG,
        ArgumentType.ULONG,
        ArgumentType.SIZE_T,
        ArgumentType.SSIZE_T,
        ArgumentType.POINTER,
    }

    _eight_byte_types = {
        ArgumentType.LONGLONG,
        ArgumentType.ULONGLONG,
    }

    _four_byte_arg_regs = [
        "r3",
        "r4",
        "r5",
        "r6",
        "r7",
        "r8",
        "r9",
        "r10",
    ]

    _eight_byte_arg_regs = [
        "r3",
        "r4",
        "r5",
        "r6",
        "r7",
        "r8",
        "r9",
        "r10",
    ]

    _soft_float = False
    _variadic_soft_float = False
    _floats_are_doubles = True
    _float_arg_regs = ["f1", "f2", "f3", "f4", "f5", "f6"]
    _double_arg_regs = ["f1", "f2", "f3", "f4", "f5", "f6"]

    _init_stack_offset = 8
    _align_stack = True
    _eight_byte_reg_size = 2
    _double_reg_size = 1
    _four_byte_stack_size = 4
    _eight_byte_stack_size = 8
    _float_stack_size = 4
    _double_stack_size = 8

    def _return_4_byte(self, emulator: emulators.Emulator, val: int) -> None:
        emulator.write_register("r3", val & self._int_inv_mask)

    def _return_8_byte(self, emulator: emulators.Emulator, val: int) -> None:
        lo = val & self._int_inv_mask
        hi = val >> 32 & self._int_inv_mask

        emulator.write_register("r3", hi)
        emulator.write_register("r4", lo)

    def _return_float(self, emulator: emulators.Emulator, val: float) -> None:
        # NOTE: powerpc promotes floats to doubles
        data = struct.pack("<d", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("f1", intval)

    def _return_double(self, emulator: emulators.Emulator, val: float) -> None:
        data = struct.pack("<d", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("f1", intval)



from .... import emulators
from ..cstd import CStdModel


class Acos(CStdModel):
    name = "acos"

    def model(self, emulator: emulators.Emulator) -> None:
        # double acos(double x);
        raise NotImplementedError()


class Acosf(CStdModel):
    name = "acosf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float acosf(float x);
        raise NotImplementedError()


class Acosl(CStdModel):
    name = "acosl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double acosl(long double x);
        raise NotImplementedError()


class Asin(CStdModel):
    name = "asin"

    def model(self, emulator: emulators.Emulator) -> None:
        # double asin(double x);
        raise NotImplementedError()


class Asinf(CStdModel):
    name = "asinf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float asinf(float x);
        raise NotImplementedError()


class Asinl(CStdModel):
    name = "asinl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double asinl(long double x);
        raise NotImplementedError()


class Asinh(CStdModel):
    name = "asinh"

    def model(self, emulator: emulators.Emulator) -> None:
        # double asinh(double x);
        raise NotImplementedError()


class Asinhf(CStdModel):
    name = "asinhf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float asinhf(float x);
        raise NotImplementedError()


class Asinhl(CStdModel):
    name = "asinhl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double asinhl(long double x);
        raise NotImplementedError()


class Atan(CStdModel):
    name = "atan"

    def model(self, emulator: emulators.Emulator) -> None:
        # double atan(double x);
        raise NotImplementedError()


class Atanf(CStdModel):
    name = "atanf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float atanf(double x);
        raise NotImplementedError()


class Atanl(CStdModel):
    name = "atanl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double atanl(long double x);
        raise NotImplementedError()


class Atan2(CStdModel):
    name = "atan2"

    def model(self, emulator: emulators.Emulator) -> None:
        # double atan2(double x, double y);
        raise NotImplementedError()


class Atan2f(CStdModel):
    name = "atan2f"

    def model(self, emulator: emulators.Emulator) -> None:
        # float atan2(float x, float y);
        raise NotImplementedError()


class Atan2l(CStdModel):
    name = "atan2l"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double atan2l(long double x, long double y);
        raise NotImplementedError()


class Atanh(CStdModel):
    name = "atanh"

    def model(self, emulator: emulators.Emulator) -> None:
        # double atanh(double x);
        raise NotImplementedError()


class Atanhf(CStdModel):
    name = "atanhf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float atanhf(float x);
        raise NotImplementedError()


class Atanhl(CStdModel):
    name = "atanhl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double atanhl(long double x);
        raise NotImplementedError()


class Cbrt(CStdModel):
    name = "cbrt"

    def model(self, emulator: emulators.Emulator) -> None:
        # double cbrt(double x);
        raise NotImplementedError()


class Cbrtf(CStdModel):
    name = "cbrtf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float cbrtf(float x);
        raise NotImplementedError()


class Cbrtl(CStdModel):
    name = "cbrtl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double long cbrtl(double x);
        raise NotImplementedError()


class Ceil(CStdModel):
    name = "ceil"

    def model(self, emulator: emulators.Emulator) -> None:
        # double ceil(double x);
        raise NotImplementedError()


class Ceilf(CStdModel):
    name = "ceilf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float ceilf(float x);
        raise NotImplementedError()


class Ceill(CStdModel):
    name = "ceill"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double ceill(long double x);
        raise NotImplementedError()


class Copysign(CStdModel):
    name = "copysign"

    def model(self, emulator: emulators.Emulator) -> None:
        # double copysign(double x, double y);
        raise NotImplementedError()


class Copysignf(CStdModel):
    name = "copysignf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float copysignf(float x, float y);
        raise NotImplementedError()


class Copysignl(CStdModel):
    name = "copysignl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double copysignl(long double x, long double y);
        raise NotImplementedError()


class Cos(CStdModel):
    name = "cos"

    def model(self, emulator: emulators.Emulator) -> None:
        # double cos(double x);
        raise NotImplementedError()


class Cosf(CStdModel):
    name = "cosf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float cosf(float x);
        raise NotImplementedError()


class Cosl(CStdModel):
    name = "cosl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double cosl(long double x);
        raise NotImplementedError()


class Cosh(CStdModel):
    name = "cosh"

    def model(self, emulator: emulators.Emulator) -> None:
        # double cosh(double x);
        raise NotImplementedError()


class Coshf(CStdModel):
    name = "coshf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float coshf(float x);
        raise NotImplementedError()


class Coshl(CStdModel):
    name = "coshl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double coshl(long double x);
        raise NotImplementedError()


class Exp(CStdModel):
    name = "exp"

    def model(self, emulator: emulators.Emulator) -> None:
        # double exp(double x);
        raise NotImplementedError()


class Expf(CStdModel):
    name = "expf"

    def model(self, emulator: emulators.Emulator) -> None:
        # float expf(float x);
        raise NotImplementedError()


class Expl(CStdModel):
    name = "expl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double expl(long double x);
        raise NotImplementedError()


class Exp2(CStdModel):
    name = "exp2"

    def model(self, emulator: emulators.Emulator) -> None:
        # double exp2(double x);
        raise NotImplementedError()


class Exp2f(CStdModel):
    name = "exp2f"

    def model(self, emulator: emulators.Emulator) -> None:
        # float exp2f(float x);
        raise NotImplementedError()


class Exp2l(CStdModel):
    name = "exp2l"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double exp2l(long double x);
        raise NotImplementedError()


class Expm1(CStdModel):
    name = "expm1"

    def model(self, emulator: emulators.Emulator) -> None:
        # double expm1(double x);
        raise NotImplementedError()


class Expm1f(CStdModel):
    name = "expm1f"

    def model(self, emulator: emulators.Emulator) -> None:
        # float expm1f(float x);
        raise NotImplementedError()


class Expm1l(CStdModel):
    name = "expm1l"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double expm1l(long double x);
        raise NotImplementedError()


class Erf(CStdModel):
    name = "erf"

    def model(self, emulator: emulators.Emulator) -> None:
        # double erf(double x);
        raise NotImplementedError()


class Erff(CStdModel):
    name = "erff"

    def model(self, emulator: emulators.Emulator) -> None:
        # float erff(float x);
        raise NotImplementedError()


class Erfl(CStdModel):
    name = "erfl"

    def model(self, emulator: emulators.Emulator) -> None:
        # long double erfl(long double x);
        raise NotImplementedError()



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import AArch64SysVModel


class AArch64SysVAbs(Abs, AArch64SysVModel):
    pass


class AArch64SysVLAbs(LAbs, AArch64SysVModel):
    pass


class AArch64SysVLLAbs(LLAbs, AArch64SysVModel):
    pass


class AArch64SysVAtof(Atof, AArch64SysVModel):
    pass


class AArch64SysVAtoi(Atoi, AArch64SysVModel):
    pass


class AArch64SysVAtol(Atol, AArch64SysVModel):
    pass


class AArch64SysVAtoll(Atoll, AArch64SysVModel):
    pass


class AArch64SysVCalloc(Calloc, AArch64SysVModel):
    pass


class AArch64SysVDiv(Div, AArch64SysVModel):
    pass


class AArch64SysVLDiv(LDiv, AArch64SysVModel):
    pass


class AArch64SysVLLDiv(LLDiv, AArch64SysVModel):
    pass


class AArch64SysVExit(Exit, AArch64SysVModel):
    pass


class AArch64SysVFree(Free, AArch64SysVModel):
    pass


class AArch64SysVMalloc(Malloc, AArch64SysVModel):
    pass


class AArch64SysVQSort(QSort, AArch64SysVModel):
    pass


class AArch64SysVRand(Rand, AArch64SysVModel):
    pass


class AArch64SysVRealloc(Realloc, AArch64SysVModel):
    pass


class AArch64SysVSrand(Srand, AArch64SysVModel):
    pass


class AArch64SysVAbort(Abort, AArch64SysVModel):
    pass


class AArch64SysVAtexit(Atexit, AArch64SysVModel):
    pass


class AArch64SysVGetenv(Getenv, AArch64SysVModel):
    pass


class AArch64SysVMblen(Mblen, AArch64SysVModel):
    pass


class AArch64SysVMbstowcs(Mbstowcs, AArch64SysVModel):
    pass


class AArch64SysVMbtowc(Mbtowc, AArch64SysVModel):
    pass


class AArch64SysVSystem(System, AArch64SysVModel):
    pass


class AArch64SysVWcstombs(Wcstombs, AArch64SysVModel):
    pass


class AArch64SysVWctomb(Wctomb, AArch64SysVModel):
    pass


__all__ = [
    "AArch64SysVAbs",
    "AArch64SysVLAbs",
    "AArch64SysVLLAbs",
    "AArch64SysVAtof",
    "AArch64SysVAtoi",
    "AArch64SysVAtol",
    "AArch64SysVAtoll",
    "AArch64SysVCalloc",
    "AArch64SysVDiv",
    "AArch64SysVLDiv",
    "AArch64SysVLLDiv",
    "AArch64SysVExit",
    "AArch64SysVFree",
    "AArch64SysVMalloc",
    "AArch64SysVQSort",
    "AArch64SysVRand",
    "AArch64SysVRealloc",
    "AArch64SysVSrand",
    "AArch64SysVAbort",
    "AArch64SysVAtexit",
    "AArch64SysVGetenv",
    "AArch64SysVMblen",
    "AArch64SysVMbstowcs",
    "AArch64SysVMbtowc",
    "AArch64SysVSystem",
    "AArch64SysVWcstombs",
    "AArch64SysVWctomb",
]



import enum
from dataclasses import dataclass


class Architecture(enum.Enum):
    """Architecture and mode.

    Names take the form ``{ARCH}_{MODE}`` where ``{ARCH}`` and ``{MODE}``
    match the regex ``[A-Z0-9]+``.

    Values are lowercase, kebab-case.
    """

    X86_32 = "x86-32"
    """32-bit x86 supporting SSE extensions."""

    X86_64 = "x86-64"
    """64-bit x86 supporting AVX2 extensions."""

    X86_64_AVX512 = "x86-64-avx512"
    """64-bit x86 supporting AVX512 extensions."""

    AARCH64 = "aarch64"

    MIPS32 = "mips32"

    MIPS64 = "mips64"

    POWERPC64 = "powerpc64"
    POWERPC32 = "powerpc32"

    ARM_V5T = "arm-v5t"
    ARM_V6M = "arm-v6m"
    ARM_V6M_THUMB = "arm-v6m-thumb"
    ARM_V7M = "arm-v7m"
    ARM_V7R = "arm-v7r"
    ARM_V7A = "arm-v7a"

    LOONGARCH32 = "loongarch32"
    LOONGARCH64 = "loongarch64"

    RISCV64 = "riscv-64"

    XTENSA = "xtensa"


class Byteorder(enum.Enum):
    """Endianness."""

    BIG = "big"
    """Big endian - most significant bit first."""

    LITTLE = "little"
    """Little endian - least significant bit first."""


class ABI(enum.Enum):
    """Application binary interface.

    Names match the regex ``[A-Z]+``.

    Values are lowercase, kebab-case.
    """

    SYSTEMV = "system-v"
    """System V."""

    CDECL = "cdecl"
    """Microsoft cdecl."""

    VECTORCALL = "vectorcall"
    """Microsoft vectorcall."""

    FASTCALL = "fastcall"
    """Microsoft fastcall."""

    NONE = "none"
    """No ABI"""


@dataclass(frozen=True)
class Platform:
    """Platform metadata/configuration storage class."""

    architecture: Architecture
    """Architecture and mode."""

    byteorder: Byteorder
    """Endianness."""

    def __repr__(self) -> str:
        return f"{self.architecture}:{self.byteorder}"


__all__ = ["Platform", "Architecture", "Byteorder", "ABI"]
